/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 *
 * This code is a derivative work of other copyrighted software.  See the
 * file COPYRIGHTS for more information on terms and conditions.
 */

extern "C" {
#include <stk.h>
}
#include <narray.hh>
#include "morphology.hh"

extern "C" PRIMITIVE
narray_bdilate(SCM image, SCM kern)
{
  Num_Array *c_image = narray_instance(image);
  Num_Array *c_kern  = narray_instance(kern);
  Num_Array *c_result = NULL;

  T *ddata, *kdata, *rdata;

  SCM result = narray(new Num_Array(c_image->dimension_ref(),
				    c_image->shape_ref()));
  c_result = narray_instance(result);

  ddata = (c_image->convension_ref() == C ? c_image->storage_iref()
	   : c_image->storage_ref(C));
  kdata = (c_kern->convension_ref() == C ? c_kern->storage_iref()
	   : c_kern->storage_ref(C));
  rdata = c_result->storage_iref();

  binary_dilate_reflected(rdata, 
			  ddata, c_image->shape_ref(0), c_image->shape_ref(1),
			  kdata, c_kern->shape_ref(0),  c_kern->shape_ref(1));

  return result;
}

extern "C" PRIMITIVE
narray_berode(SCM image, SCM kern)
{
  Num_Array *c_image = narray_instance(image);
  Num_Array *c_kern  = narray_instance(kern);
  Num_Array *c_result = NULL;

  T *ddata, *kdata, *rdata;

  SCM result = narray(new Num_Array(c_image->dimension_ref(),
				    c_image->shape_ref()));
  c_result = narray_instance(result);

  ddata = (c_image->convension_ref() == C ? c_image->storage_iref()
	   : c_image->storage_ref(C));
  kdata = (c_kern->convension_ref() == C ? c_kern->storage_iref()
	   : c_kern->storage_ref(C));
  rdata = c_result->storage_iref();

  binary_erode_reflected(rdata, 
			 ddata, c_image->shape_ref(0), c_image->shape_ref(1),
			 kdata, c_kern->shape_ref(0),  c_kern->shape_ref(1));
  
  return result;
}

extern "C" PRIMITIVE
narray_gdilate(SCM image, SCM kern)
{
  Num_Array *c_image = narray_instance(image);
  Num_Array *c_kern  = narray_instance(kern);
  Num_Array *c_result = NULL;

  T *ddata, *kdata, *rdata;

  SCM result = narray(new Num_Array(c_image->dimension_ref(),
				    c_image->shape_ref()));
  c_result = narray_instance(result);

  ddata = (c_image->convension_ref() == C ? c_image->storage_iref()
	   : c_image->storage_ref(C));
  kdata = (c_kern->convension_ref() == C ? c_kern->storage_iref()
	   : c_kern->storage_ref(C));
  rdata = c_result->storage_iref();

  grey_dilate_reflected(rdata, 
			ddata, c_image->shape_ref(0), c_image->shape_ref(1),
			kdata, c_kern->shape_ref(0),  c_kern->shape_ref(1));
  
  return result;
}

extern "C" PRIMITIVE
narray_gerode(SCM image, SCM kern)
{
  Num_Array *c_image = narray_instance(image);
  Num_Array *c_kern  = narray_instance(kern);
  Num_Array *c_result = NULL;

  T *ddata, *kdata, *rdata;

  SCM result = narray(new Num_Array(c_image->dimension_ref(),
				    c_image->shape_ref()));
  c_result = narray_instance(result);

  ddata = (c_image->convension_ref() == C ? c_image->storage_iref()
	   : c_image->storage_ref(C));
  kdata = (c_kern->convension_ref() == C ? c_kern->storage_iref()
	   : c_kern->storage_ref(C));
  rdata = c_result->storage_iref();

  grey_erode_reflected(rdata, 
		       ddata, c_image->shape_ref(0), c_image->shape_ref(1),
		       kdata, c_kern->shape_ref(0),  c_kern->shape_ref(1));
  
  return result;
}

extern "C" PRIMITIVE
narray_inf(SCM a, SCM b)
{
  Num_Array *A, *B;
  T *A_data, *B_data;

  A = narray_instance(a);
  B = narray_instance(b);
  SCM result = narray(new Num_Array(A->dimension_ref(), A->shape_ref()));

  A_data = (A->convension_ref() == C ? A->storage_iref() : A->storage_ref(C));
  B_data = (B->convension_ref() == C ? B->storage_iref() : B->storage_ref(C));
  elementwise_infimum(narray_instance(result)->storage_iref(),
		      A_data, B_data, A->size_ref());

  return result;
}

extern "C" PRIMITIVE
narray_sup(SCM a, SCM b)
{
  Num_Array *A, *B;
  T *A_data, *B_data;

  A = narray_instance(a);
  B = narray_instance(b);
  SCM result = narray(new Num_Array(A->dimension_ref(), A->shape_ref()));

  A_data = (A->convension_ref() == C ? A->storage_iref() : A->storage_ref(C));
  B_data = (B->convension_ref() == C ? B->storage_iref() : B->storage_ref(C));
  elementwise_supremum(narray_instance(result)->storage_iref(),
		       A_data, B_data, A->size_ref());

  return result;
}


extern "C" PRIMITIVE
narray_berode2(SCM image, SCM kern)
{
  Num_Array *c_image = narray_instance(image);
  Num_Array *c_kern  = narray_instance(kern);
  Num_Array *c_result = NULL;

  T *ddata, *kdata, *rdata;

  SCM result = narray(new Num_Array(c_image->dimension_ref(),
				    c_image->shape_ref()));
  c_result = narray_instance(result);

  ddata = (c_image->convension_ref() == C ? c_image->storage_iref()
	   : c_image->storage_ref(C));
  kdata = (c_kern->convension_ref() == C ? c_kern->storage_iref()
	   : c_kern->storage_ref(C));
  rdata = c_result->storage_iref();

  binary_erode_reflected2(rdata, 
			 ddata, c_image->shape_ref(0), c_image->shape_ref(1),
			 kdata, c_kern->shape_ref(0),  c_kern->shape_ref(1));
  
  return result;
}

