/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

#include "Num_Array.hh"

//-----------------------------------------------------------------------------
// Abstract Array operations defined here:
//
_Num_Array& 
_Num_Array::operator*=(T scalar)
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	*(elem_iref(ind)) *= scalar;
        inc_index(ind);
    }
    return *this;
}

_Num_Array& 
_Num_Array::operator*=(const _Num_Array& other) // NOT A MATRIX MULTIPLY!!!
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	*(elem_iref(ind)) *= other.elem_ref(ind);
        inc_index(ind);
    }
    return *this;
}

_Num_Array& 
_Num_Array::operator+=(T scalar)
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	*(elem_iref(ind)) += scalar;
        inc_index(ind);
    }
    return *this;
}

_Num_Array& 
_Num_Array::operator+=(const _Num_Array& other)
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	*(elem_iref(ind)) += other.elem_ref(ind);
        inc_index(ind);
    }
    return *this;
}

_Num_Array& 
_Num_Array::operator-=(T scalar)
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	*(elem_iref(ind)) -= scalar;
        inc_index(ind);
    }
    return *this;
}

_Num_Array& 
_Num_Array::operator-=(const _Num_Array& other)
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	*(elem_iref(ind)) -= other.elem_ref(ind);
        inc_index(ind);
    }
    return *this;
}

_Num_Array& 
_Num_Array::operator/=(T scalar)
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	*(elem_iref(ind)) /= scalar;
        inc_index(ind);
    }
    return *this;
}

void 
_Num_Array::fill(T scalar)
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	*(elem_iref(ind)) = scalar;
        inc_index(ind);
    }
}

// stat_type*
// _Num_Array::statistics() const
// {
//   // mean = (1/N) * SUM xi
//   // variance = (1/(N-1)) * (SUM (xi - mean)^2)
//   // RMS = sqrt ( (SUM xi^2)/N )
//   // skew = 1/(stdev^3 * N) * { SUM (xi - mean)^3 }
//   // kurtosis = ( 1/(stdev^4 * N) * { SUM (xi - mean)^4} ) - 3
//   // Not implemented:
//   // entropy = - ( SUM (norm_hist * log2(norm_hist) ) )
//   // contrast = SUM( (pixel_value^2) * norm_hist )
//   // norm_hist =  (num  pixels  with  gray  level  x)/(total
//   //          number pixels)
//     double tmp;
//     stat_type *stats;
//     unsigned long i;
//     index_type ind[MAX_DIMENSION];

//     stats = (stat_type *)calloc(1, sizeof(stat_type));
//     stats->num_elements = size_ref();
//     memset(ind, 0, dimension_ref() * sizeof(index_type));
//     i = stats->num_elements;

//     stats->min = stats->max = elem_ref(ind);
//     while(i-- > 0) {
// 	tmp = elem_ref(ind);
// 	stats->integral += tmp;
// 	if(tmp != 0) {
// 	    if(tmp > 0) {
// 		stats->pos_integral += tmp;
// 		stats->num_positive++;
// 	    }
// 	    else {
// 		stats->neg_integral += tmp;
// 		stats->num_negative++;
// 	    }
// 	    stats->num_nonzero++;
// 	}
// 	if(tmp < stats->min) {
// 	    stats->min = tmp;
// 	    memcpy(stats->min_position, ind, 
// 		   dimension_ref() * sizeof(index_type));
// 	}
// 	if(tmp > stats->max) {
// 	    stats->max = tmp;
// 	    memcpy(stats->max_position, ind, 
// 		   dimension_ref() * sizeof(index_type));
// 	}
// 	stats->rms += tmp * tmp;
// 	inc_index(ind);
//     }
//     stats->mean = stats->integral / stats->num_elements;

//     i = stats->num_elements;
//     memset(ind, 0, dimension_ref() * sizeof(index_type));
//     double diff;
//     while(i-- > 0) {
// 	tmp = elem_ref(ind);
// 	diff = tmp - stats->mean;
// 	tmp = diff * diff;
// 	stats->variance += tmp;
// 	tmp *= diff;
// 	stats->skewness += tmp;
// 	tmp *= diff;
// 	stats->kurtosis += tmp;
// 	inc_index(ind);
//     }

//     stats->variance /= (stats->num_elements - 1);
//     stats->stddev = sqrt(stats->variance);
//     if(stats->variance != 0.0) {
// 	stats->skewness /= stats->num_elements * pow(stats->stddev, 3.0);
// 	stats->kurtosis /= stats->num_elements * pow(stats->variance, 2.0);
// 	stats->kurtosis -= 3.0;
//     }
//     stats->rms = sqrt(stats->rms/stats->num_elements);
    
//     return stats;
// }

stat_type*
_Num_Array::stats_1(const _Num_Array *mask) const   // First order statistics
{
    double tmp;
    stat_type *stats;
    unsigned long i;
    index_type ind[MAX_DIMENSION];

    stats = (stat_type *)calloc(1, sizeof(stat_type));
    stats->num_elements = 0;
    memset(ind, 0, dimension_ref() * sizeof(index_type));
    i = size_ref();
    
    stats->min = stats->max = elem_ref(ind);
    while(i-- > 0) {
      if(mask==NULL || mask->elem_ref(ind)!=0) {
	stats->num_elements++;
	tmp = elem_ref(ind);
	stats->integral += tmp;
	if(tmp != 0) {
	  if(tmp > 0) {
	    stats->pos_integral += tmp;
	    stats->num_positive++;
	  }
	  else {
	    stats->neg_integral += tmp;
	    stats->num_negative++;
	  }
	  stats->num_nonzero++;
	}
	if(tmp < stats->min) {
	  stats->min = tmp;
	  memcpy(stats->min_position, ind, 
		 dimension_ref() * sizeof(index_type));
	}
	if(tmp > stats->max) {
	  stats->max = tmp;
	  memcpy(stats->max_position, ind, 
		 dimension_ref() * sizeof(index_type));
	}
	stats->rms += tmp * tmp;
      }
      inc_index(ind);
    }
    stats->mean = stats->integral / stats->num_elements;
    stats->rms = sqrt(stats->rms/stats->num_elements);
    
    return stats;
}

stat_type*
_Num_Array::stats_2(double the_mean, const _Num_Array *mask) const	
{
  // higher order stats about mean
  // variance = (1/(N-1)) * (SUM (xi - mean)^2)
  // skewness = 1/(stdev^3 * N) * { SUM (xi - mean)^3 }
  // kurtosis = ( 1/(stdev^4 * N) * { SUM (xi - mean)^4} ) - 3
    double tmp, diff;
    stat_type *stats;
    unsigned long i;
    index_type ind[MAX_DIMENSION];

    stats = (stat_type *)calloc(1, sizeof(stat_type));
    stats->num_elements = 0;
    stats->mean = the_mean;

    memset(ind, 0, dimension_ref() * sizeof(index_type));

    for(i=size_ref(); i>0; i--) {
      if(mask==NULL || mask->elem_ref(ind)!=0) {
	stats->num_elements++;
	tmp = elem_ref(ind);
	diff = tmp - the_mean;
	tmp = diff * diff;
	stats->sum_square_diffs += tmp;
	tmp *= diff;
	stats->sum_cube_diffs   += tmp;
	tmp *= diff;
	stats->sum_fourth_diffs += tmp;
      }
      inc_index(ind);
    }
    stats->variance = stats->sum_square_diffs / (stats->num_elements - 1);
    stats->stddev = sqrt(stats->variance);
    if(stats->variance != 0.0) {
      // calculate skewness from sum of cubed differences (with mean)
      stats->skewness  = stats->sum_cube_diffs;
      stats->skewness /= stats->num_elements * pow(stats->stddev, 3.0);
      // calculate kurtosis from sum of differences^4
      stats->kurtosis  = stats->sum_fourth_diffs;
      stats->kurtosis /= stats->num_elements * pow(stats->variance, 2.0);
      stats->kurtosis -= 3.0;
    }
    
    return stats;
}

Real 
_Num_Array::mean() const
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();
    Real sum = 0;

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	sum += elem_ref(ind);
	inc_index(ind);
    }
    return sum / size;
}

static int
my_compare(const void *a, const void *b)
{
  Real tmp = *((Real*)a) - *((Real *)b);
  return (tmp<0 ? -1 : (tmp>0 ? 1 : 0));
}

// Real
// _Num_Array::median() const
// {
//   Num_Array cpy(*this, C, Malloc);
//   unsigned long nmemb = cpy.size_ref();
//   Real *data  = cpy.storage_iref(); 

//   qsort(data, nmemb, sizeof(Real), my_compare);
  
//   return data[(int)floor(nmemb/2)];
// }

Real 
_Num_Array::max() const
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();
    Real max, tmp;

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    max = elem_ref(ind);
    inc_index(ind);
    for(unsigned long l=1; l<size; l++) {
	tmp = elem_ref(ind);
	if(tmp > max)
	    max = tmp;
	inc_index(ind);
    }
    return max;
}
 
Real 
_Num_Array::min() const
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();
    Real min, tmp;

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    min = elem_ref(ind);
    inc_index(ind);
    for(unsigned long l=1; l<size; l++) {
	tmp = elem_ref(ind);
	if(tmp < min)
	    min = tmp;
	inc_index(ind);
    }
    return min;
}

void
_Num_Array::bound_upper(T upper_bound)
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();
    T *tmp;
    
    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	tmp = elem_iref(ind);
	if(*tmp > upper_bound)
	    *tmp = upper_bound;
	inc_index(ind);
    }
}
					    
void
_Num_Array::bound_lower(T lower_bound)
{
    index_type ind[MAX_DIMENSION];
    unsigned long size = size_ref();
    T *tmp;
    
    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    for(unsigned long l=0; l<size; l++) {
	tmp = elem_iref(ind);
	if(*tmp < lower_bound)
	    *tmp = lower_bound;
	inc_index(ind);
    }
}

void
_Num_Array::normalise(T lower_bound, T upper_bound)
{
  T the_min, the_max, the_offset, the_scale;
  index_type ind[MAX_DIMENSION];
  unsigned long size;
  T *tmp;

  size = size_ref();
  the_min = min();		// Should call these min_ref(), max_ref()!!!
  the_max = max();
  the_offset = lower_bound - the_min;
  the_scale  = ((min==max==0) ? 0 
		: ((fabs(lower_bound) + fabs(upper_bound))
		   / (fabs(the_min) + fabs(the_max))));

  memset(ind,0,MAX_DIMENSION*sizeof(index_type));
  for(unsigned long l=0; l<size; l++) {
    tmp = elem_iref(ind);
    (*tmp) += the_offset;
    (*tmp) *= the_scale;
    inc_index(ind);
  }
}

void
_Num_Array::threshold(T threshold, threshold_type type, T zero, T one)
{
  index_type ind[MAX_DIMENSION];
  unsigned long size;
  T *tmp;

  memset(ind,0,MAX_DIMENSION*sizeof(index_type));
  size = size_ref();

  switch(type) {
  case Binary:		// uses real value rather than magnitude
    for(; size>0; size--) {
      tmp = elem_iref(ind);
      *tmp = ((*tmp>threshold) ? one : zero);
      inc_index(ind);
    }
    break;
  case Magnitude:	// Thresholds the magnitude (ignoring the sign)
    for(; size>0; size--) {
      tmp = elem_iref(ind);
      *tmp = (((*tmp>0 && *tmp>threshold) || (*tmp<0 && *tmp<-threshold)) 
	      ? one : zero);
      inc_index(ind);
    }
    break;
  case Hard:		// Zeros elements with magnitude less than threshold
    for(; size>0; size--) {
      tmp = elem_iref(ind);
      if(*tmp > 0) {
	if(*tmp < threshold)
	  *tmp = 0;
      }
      else
	if(-*tmp < threshold)
	  *tmp = 0;
      inc_index(ind);
    }
    break;
  case Soft:		// Shrinks elements toward zero by threshold
    for(; size>0; size--) {
      tmp = elem_iref(ind);
      if(*tmp > 0)
	*tmp = ( *tmp > threshold ? *tmp - threshold : 0);
      else
	*tmp = (-*tmp > threshold ? *tmp + threshold : 0);
      inc_index(ind);
    }
    break;
  }
}
    
void
_Num_Array::threshold(const _Num_Array *thresholds, 
		      threshold_type type, T zero, T one)
{
  index_type ind[MAX_DIMENSION];
  unsigned long size;
  T *d, *t;

  memset(ind, 0, MAX_DIMENSION*sizeof(index_type));
  size = size_ref();

  switch(type) {
  case Binary:		// uses real value rather than magnitude
    for(; size>0; size--) {
      d=elem_iref(ind); t=thresholds->elem_iref(ind);
      *d = ((*d>*t) ? one : zero);
      inc_index(ind);
    }
    break;
  case Magnitude:	// Thresholds the magnitude (ignoring the sign)
      for(; size>0; size--) {
	d=elem_iref(ind); t=thresholds->elem_iref(ind);
	*d = (((*d>0 && *d>*t) || (*d<0 && *d<-*t)) 
	      ? one : zero);
	inc_index(ind);
      }
      break;
  case Hard:		// Zeros elements with magnitude less than threshold
    for(; size>0; size--) {
      d=elem_iref(ind); t=thresholds->elem_iref(ind);
      if(*d > 0) {
	if( *d < *t)
	  *d = 0;
      }
      else
	if(-*d < *t)
	  *d = 0;
      inc_index(ind);
    }
    break;
  case Soft:		// Shrinks elements toward zero by threshold
    for(; size>0; size--) {
      d=elem_iref(ind); t=thresholds->elem_iref(ind);
      if(*d > 0)
	*d = ( *d > *t ? *d - *t : 0);
      else
	*d = (-*d > *t ? *d + *t : 0);
      inc_index(ind);
    }
    break;
  }
}

//----------------------------------------------------------------------------
// Implementation dependant array operations (includes all those which use
//	C and Fortran calls taking *storage).
Num_Array& 
Num_Array::operator*=(T scalar)
{
    T* t = storage_iref();
    unsigned long nelem = size_ref();	// A little safer than num_elements
					// although its use would be acceptable
    for(unsigned long i=0; i<nelem; i++,t++)
	*t *= scalar;
    return *this;
}

Num_Array&
Num_Array::operator*=(const Num_Array& other) // NOT A MATRIX MULTIPLY!!!
{
  T *t, *o;
  unsigned long nelem;

  if(convension_ref()==other.convension_ref() && convension_ref() != Unknown) {
    t = storage_iref(); o = other.storage_iref();
    for(nelem=size_ref(); nelem>0; nelem--, t++, o++)
      *t *= *o;
  }
  else
    _Num_Array::operator*=(other);

  return *this;
}

Num_Array*
Num_Array::clone() const
{
    Num_Array *the_clone;
    
    the_clone = new Num_Array();
    the_clone->dimension = dimension;
    memcpy(the_clone->shape, shape, 
	   dimension * sizeof(index_type));
    the_clone->convension = convension;
    the_clone->allocation = allocation;
    memcpy(the_clone->multipliers, multipliers, 
	   dimension * sizeof(index_type));
    the_clone->num_elements = num_elements;
    memcpy(the_clone->short_storage, short_storage, 
	   SHORT_STORAGE_LENGTH*sizeof(T));
    switch(allocation) 
	{
	case New:	
	    the_clone->storage = new T [num_elements];
	    memcpy(the_clone->storage_iref(), storage, num_elements*sizeof(T));
	    break;
	case Malloc:	
	    the_clone->storage = (T *)malloc(num_elements*sizeof(T));
	    memcpy(the_clone->storage_iref(), storage, num_elements*sizeof(T));
	    break;
	case Static:	
	  the_clone->storage = the_clone->short_storage;
	  if(num_elements > SHORT_STORAGE_LENGTH)
	    fprintf(stderr, "Num_Array::clone: Warning: static storage size mismatch");
	  break;
	}
    return the_clone;
}

void 
Num_Array::fill(T scalar)
{
    T* t = storage;
    unsigned long nelem = num_elements;

    for(unsigned long i=0; i<nelem; i++,t++)
	*t = scalar;
}

Real 
Num_Array::mean() const
{
    T *data = storage_iref();
    unsigned long size = size_ref();
    Real sum = 0;

    for(unsigned long l=0; l<size; l++) {
	sum += *data;
	data++;
    }
    return sum / size;
}

Real 
Num_Array::sum() const
{
    T *data = storage_iref();
    unsigned long size = size_ref();
    Real sum = 0;

    for( ; size > 0; size--) {
	sum += *data;
	data++;
    }
    return sum;
}

Real
Num_Array::median() const
{
  Num_Array cpy(*this, C, Malloc);
  unsigned long nmemb = cpy.size_ref();
  Real *data  = cpy.storage_iref(); 

  qsort(data, nmemb, sizeof(Real), my_compare);
  
  return data[(int)floor(nmemb/2)];
}

Real 
Num_Array::max() const
{
    T *data = storage_iref();
    unsigned long size = size_ref();
    Real max, tmp;

    max = *data;
    data++;
    for(unsigned long l=1; l<size; l++) {
	tmp = *data;
	if(tmp > max)
	    max = tmp;
	data++;
    }
    return max;
}
 
Real 
Num_Array::min() const
{
    T *data = storage_iref();
    unsigned long size = size_ref();
    Real min, tmp;

    min = *data;
    data++;
    for(unsigned long l=1; l<size; l++) {
	tmp = *data;
	if(tmp < min)
	    min = tmp;
	data++;
    }
    return min;
}

void
Num_Array::bound_upper(T upper_bound)
{
    unsigned long size = size_ref();
    T *tmp = storage_iref();
    
    for(unsigned long l=0; l<size; l++) {
	if(*tmp > upper_bound)
	    *tmp = upper_bound;
	tmp++;
    }
}

void
Num_Array::bound_lower(T lower_bound)
{
    unsigned long size = size_ref();
    T *tmp = storage_iref();
    
    for(unsigned long l=0; l<size; l++) {
	if(*tmp < lower_bound)
	    *tmp = lower_bound;
	tmp++;
    }
}

void
Num_Array::normalise(T lower_bound, T upper_bound)
{
  T the_min, the_max, the_offset, the_scale;
  unsigned long size, l;
  T *tmp;

  size = size_ref();
  the_min = min();		// Should call these min_ref(), max_ref()!!!
  the_max = max();
  the_offset = lower_bound - the_min;
  the_scale  = ((the_min==the_max && the_min==0) ? 0 
		: (upper_bound - lower_bound) / (the_max - the_min));
// 		: ((fabs(lower_bound) + fabs(upper_bound))
// 		   / (fabs(the_min) + fabs(the_max))));

  for(l=0, tmp=storage_iref(); l<size; l++, tmp++) {
    (*tmp) += the_offset;
    (*tmp) *= the_scale;
  }
}

void
Num_Array::threshold(T threshold, threshold_type type, T zero, T one)
{
  unsigned long size, l;
  T *tmp;

  size = size_ref();

  switch(type) {
  case Binary:		// uses real value rather than magnitude
    for(l=0, tmp=storage_iref(); l<size; l++, tmp++)
      *tmp = ((*tmp>threshold) ? one : zero);
    break;
  case Magnitude:	// Thresholds the magnitude (ignoring the sign)
    for(l=0, tmp=storage_iref(); l<size; l++, tmp++)
      *tmp = (((*tmp>0 && *tmp>threshold) || (*tmp<0 && *tmp<-threshold)) 
	      ? one : zero);
    break;
  case Hard:		// Zeros elements with magnitude less than threshold
    for(l=0, tmp=storage_iref(); l<size; l++, tmp++)
      if(*tmp > 0) {
	if( *tmp < threshold)
	  *tmp = 0;
      }
      else
	if(-*tmp < threshold)
	  *tmp = 0;
    break;
  case Soft:		// Shrinks elements toward zero by threshold
    for(l=0, tmp=storage_iref(); l<size; l++, tmp++)
      if(*tmp > 0)
	*tmp = ( *tmp > threshold ? *tmp - threshold : 0);
      else
	*tmp = (-*tmp > threshold ? *tmp + threshold : 0);
    break;
  }
}

void
Num_Array::threshold(const Num_Array *thresholds, 
		     threshold_type type, T zero, T one)
{
  unsigned long size, l;
  T *d, *t;

  if(convension_ref()!=thresholds->convension_ref())
    _Num_Array::threshold(thresholds, type, zero, one);

  else {
    size = size_ref();
    d = storage_iref();
    t = thresholds->storage_iref();

    switch(type) {
    case Binary:	// uses real value rather than magnitude
      for(l=0; l<size; l++, d++, t++)
	*d = ((*d>*t) ? one : zero);
      break;
    case Magnitude:	// Thresholds the magnitude (ignoring the sign)
      for(l=0; l<size; l++, d++, t++)
	*d = (((*d>0 && *d>*t) || (*d<0 && *d<-*t)) 
	      ? one : zero);
      break;
    case Hard:		// Zeros elements with magnitude less than threshold
      for(l=0; l<size; l++, d++, t++)
	if(*d > 0) {
	  if( *d < *t)
	    *d = 0;
	}
	else
	  if(-*d < *t)
	    *d = 0;
      break;
    case Soft:		// Shrinks elements toward zero by threshold
      for(l=0; l<size; l++, d++, t++)
	if(*d > 0)
	  *d = ( *d > *t ? *d - *t : 0);
	else
	*d = (-*d > *t ? *d + *t : 0);
      break;
    }
  }
}
    
_Num_Array& 
Num_Array::Num_block_ref(const index_type *shape,
			 const index_type *origin,
			 const index_type *steps) const
{
    Num_Subarray iref(*this, shape, origin, steps);
    Num_Array*	 block = new Num_Array(iref.dimension_ref(), iref.shape_ref());
    block->copy(iref);
    return *block;
}
    
void 
Num_Array::Num_block_set(const _Num_Array& new_value,
		     const index_type *origin,
		     const index_type *steps)
{
    Num_Subarray iref(*this, new_value.shape_ref(), origin, steps);
    iref.copy(new_value);
}

Num_Subarray * 
Num_Array::Num_block_iref(const index_type *shape,
			  const index_type *origin,
			  const index_type *steps) const
{
    return (new Num_Subarray(*this, shape, origin, steps));
}


_Num_Array& 
Num_Subarray::Num_block_ref(const index_type *shape,
			    const index_type *origin,
			    const index_type *steps) const
{
    Num_Subarray iref(*this, shape, origin, steps);
    Num_Array*   block = new Num_Array(iref.dimension_ref(), iref.shape_ref());
    block->copy(iref);
    return *block;
}
    
void 
Num_Subarray::Num_block_set(const _Num_Array& new_value,
			const index_type *origin,
			const index_type *steps)
{
    ND_Subarray iref(*this, new_value.shape_ref(), origin, steps);
    iref.copy(new_value);
}

Num_Subarray * 
Num_Subarray::Num_block_iref(const index_type *shape,
			 const index_type *origin,
			 const index_type *steps) const
{
    return (new Num_Subarray(*this, shape, origin, steps));
}













