/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

#ifndef __multiply__
#define __multiply__

/*
 * matrix_multiply:  As the name suggests.  Each argument must have dimension
 *	of '2'; vectors (row or column) are indicated by one (second or first)
 *	element of the shape being '1'.
 */
void matrix_multiply(const _Num_Array *A, const _Num_Array *B, _Num_Array *C);

/*
 * scalar_multiply:  Returns the sum of the elementwise products of two
 * arrays.  This is the 'elementwise' extension of vector multiplication.
 * It is not the "true" scalar multiply operation for anything other than 
 * vectors.
 * NB. if both args are vectors he same result can be obtained using matrix 
 * multiply---the result will be a single element array however.  This
 * function then is provided mainly for convienience and efficiency.
 */
double scalar_multiply(const _Num_Array *A, const _Num_Array *B); 
double scalar_multiply(const Num_Array *A, const Num_Array *B); 

/*
 * A variation on scalar multiplication is calculation of the norm.
 */
double Lp_norm(const _Num_Array *A, int p);
double Lp_norm(const Num_Array *A, int p);

#endif /*__multiply__*/

