/*
 *
 * n a r r a y . c			-- ND numeric arrays 
 *
 * Copyright (C) 1995 Andrew Dorrell -  <andrewd@ee.uts.edu.au>
 * 
 *
 * Permission to use, copy, and/or distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that both the above copyright notice and this permission notice appear in
 * all copies and derived works.  Fees for distribution or use of this
 * software or derived works may only be charged with express written
 * permission of the copyright holder.  
 * This software is provided ``as is'' without express or implied warranty.
 *
 * This software is a derivative work of other copyrighted softwares; the
 * copyright notices of these softwares are placed in the file COPYRIGHTS
 *
 *
 *           Author: Andrew Dorrell [andrewd@ihf.uts.edu.au]
 *    Creation date: Thu Aug 31 12:25:55 EST 1995
 * Last file update: Thu Aug 31 12:25:55 EST 1995
 */

extern "C" {
#include <stk.h>
}
#include "Num_Array.hh"
#include "Ind_Array.hh"
#include "convolve.hh"
#include "multiply.hh"
#include "narray.hh"

static void 	sweep_narray(SCM na);
static void 	mark_narray(SCM na);
static void	display_narray(SCM na, SCM port, int mode);


// PRIMITIVE make_narray(SCM arglist, int argcount);
// PRIMITIVE narrayp(SCM obj);
// PRIMITIVE narray_dimension(SCM obj);
// PRIMITIVE narray_shape(SCM obj);
// PRIMITIVE narray_set(SCM arglist, int argcount);
// PRIMITIVE narray_ref(SCM arglist, int argcount);
// PRIMITIVE narray_read_raw(SCM shape, SCM port);
// PRIMITIVE narray_write_raw(SCM narray, SCM port);
// PRIMITIVE narray_map_elements(SCM arglist, int argcount);
// PRIMITIVE narray_map_blocks(SCM arglist, int argcount);

/**** Definitions for new type tc_narray ****/

int tc_narray, tc_nsubarray;

static STk_extended_scheme_type narray_type = {
    "narray",			/* name */
    0,				/* is_procp */
    mark_narray,		/* gc_mark_fct */
    sweep_narray,		/* gc_sweep_fct */
    NULL,			/* apply_fct */
    display_narray		/* display_fct */
};

enum {
    narray_truncated =0,
    narray_padded =1,
    narray_periodic =2,
    narray_reflected =3,
    narray_ND =8,
    narray_cube =16,
    narray_matrix =32,
    narray_vector =64,
    narray_subarray =128
};

typedef struct {
    union {
	class Num_Array *numarray;	//
					// This is to grow if and when the
					// Num_Array stuff is templated -
					// There will need to be a separate
					// entry for each type - int, float ...
					// (I was using C++ to avoid this!?)
					//
    } narray_as;
    unsigned char numeric_type;
    unsigned char topology;	// Anything else a client should know about...
} Scheme_narray;

#define NARRAY(x)	((Scheme_narray *) ((x)->storage_as.extension.data))
#define LNARRAY(x)	((x)->storage_as.extension.data)
#define NARRAYP(x)	(TYPEP(x, tc_narray))
#define NARRAY_NA(x)	(NARRAY(x)->narray_as.numarray)
#define NARRAY_TYPE(x)	(NARRAY(x)->numeric_type)
#define NARRAY_TOPO(x)	(NARRAY(x)->topology)
#define NARRAY_NDP(x)	(NARRAY(x)->topology & narray_ND)


/*
 * Subarrays never hit the surface - they exist internally only.  I have
 * defined them here as an extended type as they are large dynamically
 * allocated objects and need to be garbage collected if a function
 * aborts through a call the STk_err(...).
 */
#define NSUBARRAY(x)	((Num_Subarray *) ((x)->storage_as.extension.data))

static void 	
sweep_nsubarray(SCM obj)
{
    delete(NSUBARRAY(obj));
}

static void 	
mark_nsubarray(SCM na)
{
}

static STk_extended_scheme_type nsubarray_type = {
    "nsubarray",		/* name */
    0,				/* is_procp */
    mark_nsubarray,		/* gc_mark_fct */
    sweep_nsubarray,		/* gc_sweep_fct */
    NULL,			/* apply_fct */
    NULL			/* display_fct */
};

static PRIMITIVE 
nsubarray(Num_Subarray *the_subarray)
{
    SCM z;

    NEWCELL(z, tc_nsubarray);
    z->storage_as.extension.data = the_subarray;

    return z;
}


/*
 * Other useful macros
 */
#define MIN(x,y)	((x) < (y) ? (x) : (y))
#define MAX(x,y)	((x) < (y) ? (y) : (x))

/*****************************************************************************
 * The functions which do the work are defined here.  Note that these are
 *	all written in C++
 */

// see the files *Array.hh

static void 
list2index_with_check(SCM l, index_type *i, dim_type dim)
{
    dim_type dim_check;

    for(dim_check=0; !NULLP(l); l=CDR(l)) {
	if(INTEGERP(CAR(l)))
	    i[dim_check++] = INTEGER(CAR(l));
	else
	    STk_err("bad index", l);
    }
    if(dim_check != dim)
	STk_err("index has bad length", l);
}

static SCM 
index2list(index_type *ind, int len)
{
    SCM result = NIL;

    while(--len >= 0)
	result = Cons(STk_makeinteger(ind[len]), result);

    return result;
}

SCM 
narray(Num_Array *the_narray)
{
    SCM z;

    NEWCELL(z, tc_narray);
    LNARRAY(z) = (Scheme_narray *)must_malloc(sizeof(Scheme_narray));
    NARRAY_NA(z) = the_narray;
    NARRAY(z)->numeric_type = tc_flonum;  
    NARRAY_TOPO(z) = narray_ND;

    return z;
}

SCM 
narray(dim_type dim, const index_type *shape)
{
    return narray(new Num_Array(dim, shape));
}

Num_Array *
narray_instance(SCM obj)
{
    return NARRAY_NA(obj);
}


/*****************************************************************************
 * The STk wrappers are defined here.  
 */

PRIMITIVE 
make_narray(SCM arglist, int argcount)
{
  dim_type the_dimension;
  index_type the_shape[MAX_DIMENSION];
  SCM z;
  SCM shape, start = UNBOUND, step = UNBOUND;

  arglist = STk_reverse(arglist);
  switch(argcount) {
  case 3:	step  = CAR(arglist);	arglist = CDR(arglist);
  case 2:	start = CAR(arglist);	arglist = CDR(arglist);
  case 1:	shape = CAR(arglist);	arglist = CDR(arglist);	break;
  default:	STk_err("make-narray: insufficient arguments", arglist);
  }
  /*
   * Initialise the_dimension and the_shape variables, with error checking
   */
  if(STk_listp(shape) == Ntruth)
    STk_err("make-narray: bad dimension list", shape);
  for(the_dimension=0; !NULLP(shape); shape=CDR(shape)) {
      if(STk_integerp(CAR(shape)) == Ntruth)
	  STk_err("make-narray: bad dimension list", shape);
      else
	  if(INTEGER(CAR(shape)) > 0)
	      the_shape[the_dimension++] = INTEGER(CAR(shape));
  }
  if(the_dimension==0)
      STk_err("make-narray: cannot create an narray of this shape", shape);
  
  if(start != UNBOUND && STk_numberp(start) == Ntruth)
      STk_err("make-narray: fill start value is not a number", start);
  if(step  != UNBOUND && STk_numberp(step)  == Ntruth)
      STk_err("make-narray: fill step value is not a number", step); 

  /*
   * Now create the narray any initialise it as required
   */
  z = narray(the_dimension, the_shape);

  if(start != UNBOUND) {
      if(step == UNBOUND)
	  narray_instance(z)->fill((T)numeric_value(start));
      else {
	  index_type ind[MAX_DIMENSION];
	  unsigned long size = NARRAY_NA(z)->size_ref();
	  T c_start = numeric_value(start);
	  T c_step  = numeric_value(step);

	  memset(ind, 0, the_dimension * sizeof(index_type));
	  while(size-- > 0) {
	      NARRAY_NA(z)->elem_set(ind, c_start);
	      NARRAY_NA(z)->inc_index(ind);
	      c_start += c_step;
	  }
      }
  }

  return z;
}

static void 
mark_narray(SCM obj)
{
}

static void 
sweep_narray(SCM obj)
{
    delete(narray_instance(obj));
}

static void 
display_narray(SCM narray, SCM port, int mode)
{
    index_type ind[MAX_DIMENSION];
    Num_Array *the_narray;
    unsigned long size, l;

    the_narray = narray_instance(narray);
    sprintf(STk_tkbuffer, "#[narray "); 
    STk_puts(STk_tkbuffer, PORT_FILE(port));
    STk_print(narray_shape(narray), port, mode);

    memset(ind,0,MAX_DIMENSION*sizeof(index_type));
    NARRAY_NA(narray)->convension_set(C);
    
    switch(mode) {
    case WRT_MODE:
    case TK_MODE:
	sprintf(STk_tkbuffer, " <FLOAT %x>]", (int)
		the_narray->storage_iref());
	STk_puts(STk_tkbuffer, PORT_FILE(port));
	break;
    case DSP_MODE:
	size = narray_instance(narray)->size_ref();
	STk_puts(" <FLOAT", PORT_FILE(port));
	for(l=0; l<size; l++) {
	    sprintf(STk_tkbuffer, " %f", 
		    NARRAY_NA(narray)->felem_ref(ind));
	    STk_puts(STk_tkbuffer, PORT_FILE(port));
	    NARRAY_NA(narray)->inc_index(ind);
	}
	STk_puts(">]", PORT_FILE(port));
    }
}


PRIMITIVE 
narrayp(SCM obj)
{
  return NARRAYP(obj) ? Truth : Ntruth;
}

PRIMITIVE 
narray_dimension(SCM obj)
{
  if(narrayp(obj) == Ntruth)
    STk_err("dimension: not an narray", obj);
  return STk_makeinteger(narray_instance(obj)->dimension_ref());
}

PRIMITIVE 
narray_shape(SCM obj)
{
  SCM shape_list = NIL;
  const index_type *the_shape;

  if(narrayp(obj) == Ntruth)
    STk_err("narray-shape: not an narray", obj);
  the_shape = narray_instance(obj)->shape_ref();
  for(int i=((narray_instance(obj))->dimension_ref()-1); i>=0; i--)
    shape_list = Cons(STk_makeinteger(the_shape[i]), shape_list);
  return shape_list;
}

PRIMITIVE 
narray_set(SCM arglist, int argcount)
{
    SCM	scm_args   = UNBOUND,
	scm_narray = UNBOUND,
	scm_value  = UNBOUND,
	scm_index  = UNBOUND, 
	scm_shape  = UNBOUND, 
	scm_steps  = UNBOUND;
    Num_Array   *c_narray  = NULL;
    T		c_numeric_value=0;// The value in scm_value may be a number...
    Num_Array	*c_narray_value=NULL;	//	or an narray.
    index_type  c_index[MAX_DIMENSION], 
		c_shape[MAX_DIMENSION],
		c_steps[MAX_DIMENSION];
    int		dim_check  = 0;
    bool	set_region = false, set_number = false;

    /*
     * Initialise arrays
     */
    memset(c_index, 0, MAX_DIMENSION * sizeof(index_type));
    memset(c_shape, 0, MAX_DIMENSION * sizeof(index_type));
    memset(c_steps, 0, MAX_DIMENSION * sizeof(index_type));

    /*
     * Compile the arguments from the arglist
     */
    if(NARRAYP(scm_narray = CAR(arglist))) {
	c_narray = narray_instance(scm_narray);
	dim_check = c_narray->dimension_ref();
    }
    else
	STk_err("narray-set!: not an narray", scm_narray);
    
    scm_args = STk_reverse(arglist);

    switch(argcount) {

    case 5:	// Set a region with origin, shape and step
	if(STk_listp(scm_steps = CAR(scm_args)) == Truth) {
	    scm_args = CDR(scm_args);
	    list2index_with_check(scm_steps, c_steps, dim_check);
	} 
	else
	    STk_err("narray-set!: bad list for steps", arglist);
	
    case 4: // Set a region with origin and shape
	if(STk_listp(scm_shape = CAR(scm_args)) == Truth) {
	    scm_args = CDR(scm_args);
	    list2index_with_check(scm_shape, c_shape, dim_check);
	} 
	else 
	    STk_err("narray-set!: bad list for shape", arglist); 

    case 3: // Set a value with index or a region with origin
	if(STk_listp(scm_index = CAR(scm_args)) == Truth) {
	    scm_args = CDR(scm_args);
	    list2index_with_check(scm_index, c_index, dim_check);
	} 
	else
	    STk_err("narray-set!: bad list for index", arglist);
	    
    case 2: // Fill the narray from value which is a number or an narray
	if(STk_numberp(scm_value = CAR(scm_args)) == Truth) {
	    set_number = true;
	    set_region = (argcount != 3 ? true : false);

	    switch(TYPE(scm_value)) {
	    case tc_flonum : 
		c_numeric_value = (T)FLONM(scm_value); 
		break;
	    case tc_integer:
		c_numeric_value = (T)INTEGER(scm_value);
		break;
	    case tc_bignum :
		c_numeric_value = (T)STk_integer_value(scm_value);
		break;
	    default :
		STk_err("narray-set: unknown numeric type in set value", 
			arglist);
	    }
	}
	else if(narrayp(scm_value) == Truth) {
	    set_number = false;
	    set_region = true;
	    c_narray_value = narray_instance(scm_value);
	} 
	else
	    STk_err("elem-set!: bad set value", arglist);
    }
    if(set_region) {

	Num_Subarray tmp(
	    *c_narray,
	    (argcount > 3 ? c_shape : 
	     (set_number ? 
	      c_narray->shape_ref() : c_narray_value->shape_ref())),
	    (argcount > 2 ? c_index : (index_type *)NULL),
	    (argcount > 4 ? c_steps : (index_type *)NULL));

	if(!tmp.inside())
	    STk_err("narray-set!: region not inside narray", arglist);

	if(set_number)
	    tmp.fill(c_numeric_value);
	else
	    tmp.permutable_copy(*c_narray_value);
    }
    else
	c_narray->elem_set(c_index, c_numeric_value);
    return UNDEFINED;
}



PRIMITIVE 
narray_ref(SCM arglist, int argcount)
{
    SCM		scm_args   = UNBOUND,
		scm_narray = UNBOUND, 
		scm_index  = UNBOUND, 
		scm_shape  = UNBOUND, 
		scm_steps  = UNBOUND;
    Num_Array   *c_narray  = NULL;
    index_type  c_index[MAX_DIMENSION], 
		c_shape[MAX_DIMENSION],
		c_steps[MAX_DIMENSION];
    int		dim_check= 0;

    /*
     * Initialise arrays
     */
    memset(c_index, 0, MAX_DIMENSION * sizeof(index_type));
    memset(c_shape, 0, MAX_DIMENSION * sizeof(index_type));
    memset(c_steps, 0, MAX_DIMENSION * sizeof(index_type));

    /*
     * Compile the arguments from the arglist
     */
    if(NARRAYP(scm_narray = CAR(arglist))) {
	c_narray = narray_instance(scm_narray);
	dim_check = c_narray->dimension_ref();
    } else
	STk_err("narray-ref: not an narray", arglist);

    scm_args = STk_reverse(arglist);
    switch(argcount) 
	{
	case 4: 
	    if(STk_listp(scm_steps = CAR(scm_args)) == Truth) {
		scm_args = CDR(scm_args);
		list2index_with_check(scm_steps, c_steps, dim_check);
	    } else
		STk_err("elem_ref: bad list for steps", arglist);
	    
	case 3: 
	    if(STk_listp(scm_shape = CAR(scm_args)) == Truth) {
		scm_args = CDR(scm_args);
		list2index_with_check(scm_shape, c_shape, dim_check);
	    } else 
		STk_err("elem_ref: bad list for shape", arglist); 

	case 2: 
	    if(STk_listp(scm_index = CAR(scm_args)) == Truth) {
		scm_args = CDR(scm_args);
		list2index_with_check(scm_index, c_index, dim_check);
	    } else
		STk_err("elem_ref: bad list for index", arglist); 

	    break;
	    
	default:
	    STk_err("elem-set!: insufficient arguments", arglist);
	}

    if(argcount == 2) {	// Return value is a number
	if(!c_narray->inside(c_index))
	    STk_err("narray-ref: index is out-of-bounds", arglist);
	switch(NARRAY_TYPE(scm_narray)) 
	    {
	    case tc_flonum : 
		return STk_makenumber((double)c_narray->elem_ref(c_index));
	    case tc_integer:
	    case tc_bignum :
		return STk_makeinteger((long) c_narray->elem_ref(c_index));
	    default :
		STk_err("narray-ref: unknown numeric type in", scm_narray);
	    }
	
    } else {		// Return value is an narray
	/*
	 * Calculate the dimension of the referenced subarray (this is messy
	 *	and should be addressed in ND_Subarray::_init)
	 */
	Num_Subarray tmp(*c_narray, 
			 c_shape, 
			 c_index,
			 (index_type*)(scm_steps == UNBOUND ? NULL : c_steps));

	if(!tmp.inside())
	    STk_err("narray-ref: out of bounds reference", arglist);
	/*
	 * Copy the result into a new array.  This may be removed if a way
	 *	can be found to use shared arrays in STk without nasty effects
	 */
	SCM result = narray(tmp.dimension_ref(), tmp.shape_ref());
	NARRAY_NA(result)->copy(tmp);
	return (result);
    }
    return UNDEFINED;		// this just shuts up some compiler warnings
}

static PRIMITIVE
do_narray_permute(SCM array, SCM new_order, int imperitive)
{
    dim_type c_new_order[MAX_DIMENSION];
    Num_Array *c_array;
    dim_type   dim, dim_check;
    SCM l = new_order;

    if(!NARRAYP(array))
	STk_err("narray-permute: not an narray", array);

    if(STk_listp(new_order) != Truth)
	STk_err("narray-permute: not a list", new_order);

    dim = NARRAY_NA(array)->dimension_ref();

    for(dim_check=0; !NULLP(l) && dim_check<MAX_DIMENSION; l=CDR(l)) {
	if(INTEGERP(CAR(l)) && INTEGER(CAR(l)) < dim)
	    c_new_order[dim_check++] = INTEGER(CAR(l));
	else
	    STk_err("narray-permute: bad value in permutation", new_order);
    }
    if(dim_check != dim)
	STk_err("narray-permute: length of permutation not equal dimension",
		LIST2(array, new_order));

    if(imperitive)
	c_array = NARRAY_NA(array);
    else {
	c_array = new Num_Array(dim, NARRAY_NA(array)->shape_ref());
	c_array->copy(*(NARRAY_NA(array)));
    }

    c_array->permute(c_new_order);

    return (imperitive ? UNDEFINED : narray(c_array));
}
  
PRIMITIVE
narray_permute(SCM array, SCM new_order)
{
    return do_narray_permute(array, new_order, false);
}

PRIMITIVE
imperitive_narray_permute(SCM array, SCM new_order)
{
    return do_narray_permute(array, new_order, true);
}
 
/*
 * These readers and writers should really be made member functions of the
 * appropriate array classes... therefore some of the following may seem
 * out-of-place.
 */

enum byte_order_type {Native, Bigendian, Littlendian};

PRIMITIVE 
narray_read_raw(SCM shape, SCM port)
{
    SCM narray;
    unsigned long num_items;
    index_type ind[MAX_DIMENSION];
    T* elem; 

    if(port == UNBOUND)
	port = STk_current_input_port();
    if(!IPORTP(port))
	STk_err("narray-read-raw: not an input port", port);
    narray = make_narray(LIST1(shape), 1);
    num_items = narray_instance(narray)->size_ref();
    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));

    elem = NARRAY_NA(narray)->storage_iref();
    while(num_items--) {
	if(!(fread(elem, sizeof(T), 1, PORT_FILE(port))))
	    STk_err("narray-read-raw: error occurred while reading", port);
	elem++;
    }
    return narray;
}

/*
 * In support of bit manipulations:
 */
static unsigned char bit_mask[8] = 
/* {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80}; */
{0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};
#define bit_ref(vector, n) ((((vector)[(n)/8] & bit_mask[(n)%8]) == 0) ? 0 : 1)
#define bit_set(vector, n) ((vector)[(n)/8] |= bit_mask[(n)%8])
#define bit_clear(vector,n) ((vector)[(n)/8]  &= (~bit_mask[(n)%8]))


PRIMITIVE
narray_read_bits(SCM shape, SCM port)
{
  SCM narray;
  unsigned long num_items;
  T* elem;
  unsigned char c, i;
  
  if(port == UNBOUND)
    port = STk_current_input_port();
  if(!IPORTP(port))
    STk_err("narray-read-bits: not an input port", port);

  narray = make_narray(LIST1(shape), 1);
  num_items = NARRAY_NA(narray)->size_ref();
  
  elem = NARRAY_NA(narray)->storage_iref();

  /*
   * Read in the bits, unwrapping a byte at a time...
   */
  while(num_items >= 8) {
    c = (unsigned char)fgetc(PORT_FILE(port));
    *elem++ = (((c & bit_mask[0]) == 0) ? 0 : 1);
    *elem++ = (((c & bit_mask[1]) == 0) ? 0 : 1);
    *elem++ = (((c & bit_mask[2]) == 0) ? 0 : 1);
    *elem++ = (((c & bit_mask[3]) == 0) ? 0 : 1);
    *elem++ = (((c & bit_mask[4]) == 0) ? 0 : 1);
    *elem++ = (((c & bit_mask[5]) == 0) ? 0 : 1);
    *elem++ = (((c & bit_mask[6]) == 0) ? 0 : 1);
    *elem++ = (((c & bit_mask[7]) == 0) ? 0 : 1);
    num_items -= 8;
  }
  /*
   * Collect the straglers (bits at the end not filling a byte)
   */
  c = fgetc(PORT_FILE(port));
  for(i=0; num_items>0; i++,num_items--)
    *elem++ = bit_ref(&c, i);
  
  return narray;
}

PRIMITIVE 
narray_read_bytes(SCM shape, SCM port)
{
    SCM narray;
    unsigned long num_items;
    index_type ind[MAX_DIMENSION];
    unsigned char c;
    T *elem;

    if(port == UNBOUND)
	port = STk_current_input_port();
    if(!IPORTP(port))
	STk_err("narray-read-raw: not an input port", port);
    narray = make_narray(LIST1(shape), 1);
    num_items = NARRAY_NA(narray)->size_ref();
    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));

    elem = NARRAY_NA(narray)->storage_iref();
    while(num_items--) {
	if(fread(&c, 1, 1, PORT_FILE(port)) == 0)
	    STk_err("narray-read-raw: error occurred while reading", port);
	*elem++ = (T)c;
    }
    return narray;
}

PRIMITIVE
narray_read_short(SCM byte_order, SCM shape, SCM port)
{
    byte_order_type c_byte_order;
    T *elem;
    SCM  narray;
    long num_items;
    FILE *stream;
    unsigned char c[2];

    if(SYMBOLP(byte_order)) {
	if(EQ(byte_order, Intern("native"))) 
	    c_byte_order = Native;
	else if(EQ(byte_order, Intern("bigendian"))) 
	    c_byte_order = Bigendian;
	else if(EQ(byte_order, Intern("littlendian"))) 
	    c_byte_order = Littlendian;
	else 
	    STk_err("narray-read-short: byte order must be native, bigendian or littlendian", byte_order);
    }
    else 
	STk_err("narray-read-short: byte order must be native, bigendian or littlendian", byte_order);

    if(port == UNBOUND)
	port = STk_current_input_port();
    if(!IPORTP(port))
	STk_err("narray-read-raw: not an input port", port);

    narray = make_narray(LIST1(shape), 1);
    num_items = NARRAY_NA(narray)->size_ref();
    stream = PORT_FILE(port);
    elem = NARRAY_NA(narray)->storage_iref(); // we assume 'C' ordering

    switch(c_byte_order) {
    case Native:
	while(num_items--) {
	    if(fread(&elem, 1, 1, stream) != 1)
		STk_err("narray-read-short: error occurred while reading", 
			port);
	    elem++;
	}
	break;
	
    case Bigendian:
	while(num_items--) {
	    if(fread(c, 2, 1, stream) != 1)
		STk_err("narray-read-short: error occurred while reading", 
			port);
	    *elem = (((c[0] & 0xff) << 8) + (c[1] & 0xff));
	    elem++;
	}
	break;

    case Littlendian:
	while(num_items--) {
	    if(fread(c, 2, 1, stream) != 1)
		STk_err("narray-read-short: error occurred while reading", 
			port);
	    *elem = ((c[0] & 0xff) + ((c[1] & 0xff) << 8));
	    elem++;
	}
	break;
    }
    
    return narray;
}

PRIMITIVE
narray_write_bits(SCM array, SCM port)
{
  char i, bytes[4];
  int  need_free = 0;
  Num_Array *c_array;
  float *d;
  unsigned long n;

  if(NARRAYP(array)) {
    c_array=narray_instance(array);
    if(c_array->convension_ref() == C)
      d = c_array->storage_iref();
    else {
      d = NULL; //c_array->storage_ref(C);
      //need_free = 1;
    }
  }
  else
    STk_err("narray-write-bits: not an narray", array);
  
  if(port == UNBOUND)
    port = STk_current_input_port();
  if(!OPORTP(port))
    STk_err("narray-write-bytes: not a valid output port", port);

  if(d != NULL) {	// Row-wise storage ==> fast implementation
    for(n=c_array->size_ref(); n>31; n-=32){
      memset(bytes, 0, 4);
      for(i=0; i<4; i++){
	if(*d++ != 0)	bit_set(&bytes[i], 0);
	if(*d++ != 0)	bit_set(&bytes[i], 1);
	if(*d++ != 0)	bit_set(&bytes[i], 2);
	if(*d++ != 0)	bit_set(&bytes[i], 3);
	if(*d++ != 0)	bit_set(&bytes[i], 4);
	if(*d++ != 0)	bit_set(&bytes[i], 5);
	if(*d++ != 0)	bit_set(&bytes[i], 6);
	if(*d++ != 0)	bit_set(&bytes[i], 7);
      }
      fwrite(bytes, 1, 4, PORT_FILE(port));
    }
    if(n!=0){
      for(i=0,bytes[0]=0; n>0; i++, n--) {
	if(*d++ != 0)	bit_set(&bytes[0], i);
	if(i==7){
	  fwrite(bytes, 1, 1, PORT_FILE(port));
	  i=-1; bytes[0] = 0;
	}
      }
      if(i != 0)
	fwrite(bytes, 1, 1, PORT_FILE(port));
    }
  }
  else {		// Use slower but general method

    index_type ind[MAX_DIMENSION];
    int bit;

    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));
    for(n=c_array->size_ref(); n>31; n-=32){
      memset(bytes, 0, 4);
      for(i=0; i<4; i++)
	for(bit=0; bit<8; bit++) {
	  if(c_array->elem_ref(ind) != 0)
	    bit_set(&bytes[i], bit);
	  c_array->inc_index(ind);
	}
      fwrite(bytes, 1, 4, PORT_FILE(port));
    }
    if(n != 0) {
      for(i=0,bytes[0]=0; n>0; i++, n--) {
	if(c_array->elem_ref(ind) != 0)
	  bit_set(&bytes[0], i);
	if(i==7) {
	  fwrite(bytes, 1, 1, PORT_FILE(port));
	  i=-1; bytes[0] = 0;
	}
      }
      if(i != 0)
	fwrite(bytes, 1, 1, PORT_FILE(port));
    }
  }
    
//   if(need_free != 0)
//     free(d);

  return UNDEFINED;
}   

PRIMITIVE
narray_write_bytes(SCM narray, SCM port)
{
    if(!NARRAYP(narray))
	STk_err("narray-write-bytes: not an narray", narray);
    if(port == UNBOUND)
	port = STk_current_input_port();
    if(!OPORTP(port))
	STk_err("narray-write-bytes: not a valid output port", port);

    Num_Array tmp(NARRAY_NA(narray)->dimension_ref(), 
		  NARRAY_NA(narray)->shape_ref(),
		  C);
    tmp.copy(*(NARRAY_NA(narray)));
    tmp.bound_upper(255);
    tmp.bound_lower(0);
    index_type ind[MAX_DIMENSION];
    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));
    
    unsigned long size = tmp.size_ref();
    while(size--) {
	putc((char)(NARRAY_NA(narray)->felem_ref(ind)), PORT_FILE(port));
	NARRAY_NA(narray)->inc_index(ind);
    }
    
    STk_flush(port);
    return UNDEFINED;
}

PRIMITIVE 
narray_write_raw(SCM narray, SCM port)
{
    index_type ind[MAX_DIMENSION];
    unsigned long num_elements;

    if(narrayp(narray) == Ntruth)
	STk_err("narray-write-raw: not an narray", narray);
    if(port == UNBOUND)
	port = STk_current_input_port();
    if(!OPORTP(port))
	STk_err("narray-write-raw: not an output port", port);

    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));
    num_elements = NARRAY_NA(narray)->size_ref();

    while(num_elements--) {
	if(!(fwrite(narray_instance(narray)->elem_iref(ind),
		  sizeof(T), 1,
		  PORT_FILE(port))))
	    STk_err("narrar-write-raw: Error occured while writing", port);
	NARRAY_NA(narray)->inc_index(ind);
    }
    STk_flush(port);
    return UNDEFINED;
}

PRIMITIVE 
narray_write_ascii(SCM narray, SCM port)
{
    index_type ind[MAX_DIMENSION];
    const index_type *shape;
    char point_str[64];		// A bit arbitrary but should be OK.
    unsigned long l;
    dim_type d, dim;
    T tmp;

    if(narrayp(narray) == Ntruth)
	STk_err("narray-write-ascii: not an narray", narray);
    if(port == UNBOUND)
	port = STk_current_output_port();
    if(!OPORTP(port))
	STk_err("narray-write-ascii: not an output port", port);

    l     = NARRAY_NA(narray)->size_ref();
    dim   = NARRAY_NA(narray)->dimension_ref();
    shape = NARRAY_NA(narray)->shape_ref();
    memset(ind, 0, dim * sizeof(index_type));

    while( l-- > 0 ) {
	tmp = NARRAY_NA(narray)->felem_ref(ind);

	if((tmp - (int)tmp) == 0.0)
	    sprintf(point_str, "%li ", (long)tmp);
	else
	    sprintf(point_str, "%f ", tmp);
	fputs(point_str, PORT_FILE(port));

	for(d=0; d<dim-1; d++)
	    if(ind[d] == (shape[d]-1))
		fputs("\n", PORT_FILE(port));
	//	    else if(d==0)
	//		fputs("\t", PORT_FILE(port));
	
	NARRAY_NA(narray)->inc_index(ind);
    }

    STk_flush(port);
    return UNDEFINED;
}

PRIMITIVE
narray_read_ascii(SCM shape, SCM port)
{
    SCM narray;
    unsigned long num_items;
    //    index_type ind[MAX_DIMENSION];
    float *tmp;

    if(port == UNBOUND)
	port = STk_current_input_port();
    if(!IPORTP(port))
	STk_err("narray-read-raw: not an input port", port);
    narray = make_narray(LIST1(shape), 1);
    num_items = narray_instance(narray)->size_ref();
    //    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));
    tmp = narray_instance(narray)->storage_iref();

    while(num_items--) {
      if(!(fscanf(PORT_FILE(port), "%f", tmp++)))
	STk_err("narray-read-ascii: error occurred while reading", port);
      //	NARRAY_NA(narray)->elem_set(ind, (T)tmp);
      //	NARRAY_NA(narray)->inc_index(ind);
    }
    return narray;
}

PRIMITIVE
narray_read_memory(SCM hack)
{
  Num_Array *a;
  index_type shape[MAX_DIMENSION];
  char tmp[32], type[32];
  int  addr, i;
  float *f;

  if(!STRINGP(hack) ||
     sscanf(CHARS(hack), "#[%s (%ld %ld) <%s %x>]", 
	    tmp, shape, shape+1, type, &addr) != 5)
    STk_err("narray-read-memory: arg must be a string of the form \"[tag (shape) <type address>]\"",
	    hack);

  a = new Num_Array(2, shape);
  f = a->storage_iref();
  if(strcmp(type, "BYTE") == 0) {
    unsigned  char *c = (unsigned char *)addr;
    for(i=shape[0]*shape[1]; i>0; i--)
      *f++ = *c++;
  }
  else if(strcmp(type, "SHORT") == 0) {
    unsigned short *s = (unsigned short *)addr;
    for(i=shape[0]*shape[1]; i>0; i--)
      *f++ = *s++;
  }
  else if(strcmp(type, "INTEGER") == 0) {
    unsigned int *d = (unsigned int *)addr;
    for(i=shape[0]*shape[1]; i>0; i--)
      *f++ = *d++;
  }
  else if(strcmp(type, "FLOAT") == 0) {
    memcpy(f, (float *)addr, shape[0] * shape[1] * sizeof(float));
  }

  return narray(a);
}
  

static PRIMITIVE
internal_fast_map_elements(SCM proc, SCM args)
{
    SCM  result, save = args;		// save prevents garbage collection
    SCM  scm_args[MAX_DIMENSION];
    SCM  scm_arg_list = NIL;
    T   *c_args[MAX_DIMENSION];
    T   *c_result;
    unsigned long i,j, nargs = STk_llength(args);
    unsigned long nelem;

    result = narray(NARRAY_NA(CAR(args))->clone());
    c_result = NARRAY_NA(result)->storage_iref();

    for(i=0; i<nargs; i++, args=CDR(args)) {
	c_args[i]    = NARRAY_NA(CAR(args))->storage_iref();
	scm_args[i]  = STk_makenumber(*(c_args[i]));
	scm_arg_list = Cons(scm_args[i], scm_arg_list);
    }
    scm_arg_list = STk_reverse(scm_arg_list);

    nelem = NARRAY_NA(result)->size_ref();
    for(i=0; i<nelem; i++) {
	*c_result++ = (T)numeric_value(STk_apply(proc, scm_arg_list));
	for(j=0; j<nargs; j++) {
	    ++(c_args[j]);
	    FLONM(scm_args[j]) = *(c_args[j]);
	}
    }

    return result;
}

PRIMITIVE 
narray_map_elements(SCM arglist, int argcount)
{
    SCM	scm_proc, scm_nargs=NIL;// proc and its list of args (pointing to...)
    SCM	scm_n[MAX_DIMENSION];	// Storage for array elements applied to proc
    SCM scm_ref_arg=UNDEFINED;	// reference argument (the last narray arg)
    SCM	scm_result=UNDEFINED;	// result of proc --- an narray or a number
    SCM scm_tmp_result;

    index_type		c_ind[MAX_DIMENSION];
    const index_type	*c_ref_shape=NULL, *c_this_shape;
    index_type		c_result_shape[MAX_DIMENSION];
    dim_type		c_result_dimension;
    Num_Array		*c_result, *c_args[256];
    long	c_arg_size;
    int		i, j;
    SCM		arglist_save = arglist;

    if( argcount < 2 )
	STk_err( "narray-map: Insuffucient arguments", arglist );
    scm_proc = CAR( arglist );
    argcount--;
    /*
     * Build an array containing the elements of the arglist to which 'proc'
     *	will be applied.  A second array of SCM is also built and its objects
     *	placed into a list.  This list is then used in the call to apply.
     */
    if( argcount > 256 )
	STk_err( "narray_map_elememts: (internal) too many args", arglist );
    for( i=0; i < argcount; i++ ) {
	arglist = STk_reverse(CDR( arglist ));
	if(NARRAYP((scm_ref_arg = CAR( arglist )))) {
	    c_args[i] = narray_instance( CAR( arglist ));
	    scm_nargs = Cons( ( scm_n[i] = STk_makenumber( 0 ) ), scm_nargs );
	    /*
	     * Check for dimensional conformace.
	     * This should be done using a call to the member fn 'conforms'
	     */
	    if( i == 0 ) {
		c_ref_shape = c_args[i]->shape_ref();
	    } else {
		c_this_shape = c_args[i]->shape_ref();
		for( j=0; j < c_args[i]->dimension_ref(); j++ )
		    if( c_ref_shape[j] != c_this_shape[j] )
			STk_err( "narray-map-elements: array has bad shape", 
				CAR( arglist ));
	    }
	} else
	    STk_err( "narray-map-elements: not an narray", CAR( arglist )); 
    }
    scm_nargs = STk_reverse( scm_nargs );
    /*
     * The result returned by 'proc' is permitted to be either a number or
     *	an narray.  If an narray, it must return an narray of the same shape
     *	on each call.  The type and shape of the result is determined by
     *	the first application; space for the full result (an narray) is then
     *	allocated prior to applying 'proc' to the remaining elements.
     */
    memset( c_ind, 0, MAX_DIMENSION * sizeof( index_type ));
    c_arg_size = c_args[0]->size_ref();
    for(i=0; i<argcount; i++)
	FLONM( scm_n[i] ) = (double)c_args[i]->felem_ref( c_ind );
    scm_tmp_result = STk_apply( scm_proc, scm_nargs );

    if(NUMBERP(scm_tmp_result)) {
	bool same_convension = true;
	for(i=i; i<argcount; i++)
	    same_convension &= ((c_args[i-1]->convension_ref()
				 == c_args[i]->convension_ref())
				&& c_args[i]->convension_ref() != Unknown);
	if(same_convension) {
	    return internal_fast_map_elements(scm_proc, CDR(arglist_save));
	}
	else {
	    scm_result = make_narray(LIST1(narray_shape(scm_ref_arg)), 1);
	    c_result = narray_instance( scm_result );
	    c_result->elem_set(c_ind, (T)numeric_value(scm_tmp_result));
	    c_result->inc_index(c_ind);
	    
	    for( j=1; j<c_arg_size; j++ ) {
		for( i=0; i<argcount; i++ ) {
		    FLONM(scm_n[i]) = (double)c_args[i]->felem_ref(c_ind);
		}
		c_result->elem_set(c_ind, 
				   (T)numeric_value(STk_apply(scm_proc, 
							      scm_nargs)));
		c_result->inc_index(c_ind);
	    }
	}
    }
    else if(NARRAYP(scm_tmp_result)) {
	dim_type	c_elem_dimension;
	index_type      c_elem_shape[MAX_DIMENSION],
			c_finalresult_shape[MAX_DIMENSION];

	/*
	 * The first step is to calculate the shape for the final result
	 * (ie the result from the map operation).  This is formed by a
	 * concatination of all of the results produced by application of
	 * scm_proc to the blocks drawn from the narray arguments.
	 * Therefore:
	 *	dimension of result = dimension of the value returned by proc
	 *	shape of the result = a packing of the proc shape onto the
	 *		previously calculated 'result_shape'---which actually
	 *		specifies the number and packing of the elements
	 *		produced by application of proc, in the final result
	 * To produce the final result shape, the element shape and the
	 * (proc) result shape are first padded with 1s till they are the 
	 * same length:
	 */
	c_result_dimension = c_args[0]->dimension_ref();
	memcpy(c_result_shape, c_args[0]->shape_ref(), 
	       c_result_dimension * sizeof(index_type));
	c_elem_dimension = NARRAY_NA(scm_tmp_result)->dimension_ref();
	memcpy(c_elem_shape, NARRAY_NA(scm_tmp_result)->shape_ref(),
	       MAX_DIMENSION * sizeof(index_type));

	if(c_result_dimension < c_elem_dimension)
	    for(i=c_result_dimension; i<c_elem_dimension; i++)
		c_result_shape[c_result_dimension++] = 1;
	else if(c_result_dimension > c_elem_dimension)
	    for(i=c_elem_dimension; i<c_result_dimension; i++)
		c_elem_shape[i] = 1;
	/*
	 * The two shapes are then concatinated
	 */
	for(i=0; i<c_result_dimension; i++)
	    c_result_shape[c_result_dimension+i] = c_elem_shape[i];
	/*
	 * and then packed down to the required dimension
	 */
	int pstep = MAX(c_elem_dimension, c_result_dimension);
	pack_shape(pstep,				  // packing step 
		   c_result_dimension*2, c_result_shape,  // input shape & dim
		   pstep, c_finalresult_shape);		  // output shape & dim
	/*
	 * Now we have all the info we need to allocate space for the result,
	 */
	scm_result = narray(pstep, c_finalresult_shape);
	c_result = narray_instance(scm_result);

	/*
	 * and define the insertion region for the intermediate results
	 * (ie from applications of proc).  The insertion region (in result)
	 * is defined as a subarray on result to which the return value of
	 * proc is copied.
	 */
	Num_Subarray c_result_elem(*c_result, c_elem_shape);
	/*
	 * Now copy the first result
	 */
	c_result_elem.copy(*(NARRAY_NA(scm_tmp_result)));
	/*
	 * Then calculate and insert the remaining results
	 */
	for(j=1; j<c_arg_size; j++) {
	    c_result_elem.inc_position(c_elem_shape);
	    c_args[0]->inc_index(c_ind);
	    for(i=0; i<argcount; i++)
		FLONM(scm_n[i]) = (double)c_args[i]->felem_ref(c_ind);
	    c_result_elem.copy(*(NARRAY_NA(STk_apply(scm_proc, scm_nargs))));
	}
    }
    else
	STk_err("narray-map-blocks: proc must return a number or narray",
		scm_proc);

    return scm_result;
}

/*
 * Call for (narray-map-blocks proc shape steps narray1...)
 */
PRIMITIVE 
narray_map_blocks(SCM arglist, int argcount)
{
    SCM scm_result=UNDEFINED, scm_tmp_result;
    SCM scm_arg_shape, scm_arg_steps;
    SCM scm_proc, scm_args=NIL;
    SCM scm_arg_array[256], scm_subarrays[256];

    Num_Array  *c_result = NULL;
    Num_Subarray *c_subarray[MAX_DIMENSION];
    Num_Array *c_ref_array;
    index_type c_arg_shape[MAX_DIMENSION], c_arg_steps[MAX_DIMENSION],
	       c_result_shape[MAX_DIMENSION];
    dim_type   c_result_dimension;
    index_type k;
    long i,j;

    dim_type	c_elem_dimension;
    index_type	c_elem_shape[MAX_DIMENSION],
		c_finalresult_shape[MAX_DIMENSION];

    if( argcount < 4 )
	STk_err( "narray-map: Insuffucient arguments", arglist );

    scm_proc = CAR(arglist);
    arglist = CDR(arglist);
    argcount--;

    scm_arg_shape = CAR(arglist);
    arglist = CDR(arglist);
    argcount--;

    scm_arg_steps = CAR(arglist);
    arglist = CDR(arglist);
    argcount--;

    /*
     * compile the narray args.  The list of arguments for proc is created
     * only once and on each iteration, new data is copied to each arg.  To
     * streamline this, a scm_arg_array is stored beside the scm_args (list).
     * Each list item can be accessed from the array in the same order.
     *
     * To reference each block of the input narrays, an array of subarrays
     * is also created. On each iteration, the subarray position is 
     * incremented as requested by scm_arg_steps and the data copied to the
     * appropriate scm_arg_array entry (and implicitly to scm_args) before
     * the call to apply(scm_proc, scm_args).
     */
    if(argcount > 256)
	STk_err("narray-map-blocks: Internal error - too many args", arglist);
    SCM this_narray;
    /*
     * Get reference parameters from the first narray argument and check.
     * If all is OK then add it to the list of args before looping through
     * any remaining narray args.
     */
    if(!NARRAYP((this_narray = CAR(arglist))))
	STk_err("narray-map-blocks: not an narray", this_narray);
    
    /*
     * OK we have an narray
     */
    c_ref_array = narray_instance(this_narray);
    list2index_with_check(scm_arg_shape, c_arg_shape, 
			  c_ref_array->dimension_ref());
    list2index_with_check(scm_arg_steps, c_arg_steps,
			  c_ref_array->dimension_ref());
    /*
     * and the shape and steps args are of the right type and length.
     * Now check that they "fit" into the narray args.
     */
    for(i=0,c_result_dimension=0; i<c_ref_array->dimension_ref(); i++) {
	k = c_ref_array->shape_ref(i);
	if(c_arg_steps[i]==0 && k==c_arg_shape[i])
	    c_result_shape[c_result_dimension++] = 1;

	else if(c_arg_steps[i]==c_arg_shape[i] && !(k % c_arg_shape[i]))
	    c_result_shape[c_result_dimension++] = k / c_arg_shape[i];

	else if((k-c_arg_shape[i]) % c_arg_steps[i] == 0)
	    c_result_shape[c_result_dimension++] = 
		(k-c_arg_shape[i]) / c_arg_steps[i] + 1;

	else if(c_arg_steps[i] > c_arg_shape[i] && !(k % c_arg_steps[i]))
	  c_result_shape[c_result_dimension++] = k / c_arg_steps[i];
	
	else
	    STk_err("narray-map-blocks: shape doesn't fit into arg shape",
		    LIST3(scm_arg_shape, scm_arg_steps, this_narray));
    }
    /*
     * All being well, lets start building the arg list.  Note that
     * even the subarrays are placed into SCM objects so that they may
     * be cleaned up by the garbage collector later on. We MUST do this
     *	scm_args	is a list of narray objects to pass to proc
     *  scm_subarrays	is an array of subarray references to the
     *			narray objects passed to narray-map-blocks 
     */
    scm_args = Cons((scm_arg_array[0] = 
		     make_narray(LIST1(scm_arg_shape), 1)), 
		    scm_args);
    scm_subarrays[0] = 
	nsubarray((c_subarray[0] = 
		   new Num_Subarray(*c_ref_array, c_arg_shape)));
    arglist = CDR(arglist);
    /*
     * Now do the same for any remaining narray args in arglist
     */
    for(i=1; i<argcount; i++) {
	if(NARRAYP((this_narray = CAR(arglist)))) {
	    if(narray_instance(this_narray)->conforms(*c_ref_array)) {
		scm_args = Cons((scm_arg_array[i] = 
				 make_narray(LIST1(scm_arg_shape), 1)), 
				scm_args);
		scm_subarrays[i] = nsubarray(
		    (c_subarray[i] = 
		     new Num_Subarray(*(narray_instance(this_narray)),
				      c_arg_shape)));
		arglist = CDR(arglist);
	    }
	    else	// This narray is the wrong shape
		STk_err("narray-map-blocks: non-conformant narray",
			this_narray);
	}
	else	// One of the arguments is not an narray - abort!
	    STk_err("narray-map-blocks: not an narray", this_narray);
    }
    /* 
     * Oops! we built the scm_arg list in reverse order to the actual
     * argument order.  The quick fix (and a more complex one is not
     * really necessary) is just to reverse the list:
     */
    scm_args = STk_reverse(scm_args);
    /*
     * We now have the arguments compiled. The first step is to apply proc
     * once to determine the type and size of its result.  Space is then
     * allocated for the overall result based on this before proc is applied
     * to successive input blocks.  */
    for(i=0; i<argcount; i++)
	NARRAY_NA(scm_arg_array[i])->permutable_copy(*c_subarray[i]);
    scm_tmp_result = STk_apply(scm_proc, scm_args);
    if(NUMBERP(scm_tmp_result)) {
	long size;
	index_type ind[MAX_DIMENSION];
	memset(ind, 0, MAX_DIMENSION*sizeof(index_type));
	scm_result = narray(c_result_dimension, c_result_shape);
	c_result = narray_instance(scm_result);
	c_result->elem_set(ind, (T)numeric_value(scm_tmp_result));

	size = c_result->size_ref();
	for(j=1; j<size; j++) {
	    c_result->inc_index(ind);
	    for(i=0; i<argcount; i++) {
		c_subarray[i]->inc_position(c_arg_steps);
		NARRAY_NA(scm_arg_array[i])->permutable_copy(*c_subarray[i]);
	    }
	    c_result->elem_set(ind, (T)numeric_value(
		STk_apply(scm_proc, scm_args)));
	}
    }
    else if(NARRAYP(scm_tmp_result)) {

	c_elem_dimension = NARRAY_NA(scm_tmp_result)->dimension_ref();
	memcpy(c_elem_shape, NARRAY_NA(scm_tmp_result)->shape_ref(),
	       MAX_DIMENSION * sizeof(index_type));

	/*
	 * The first step is to calculate the shape for the final result
	 * (ie the result from the map operation).  This is formed by a
	 * concatination of all of the results produced by application of
	 * scm_proc to the blocks drawn from the narray arguments.
	 * Therefore:
	 *	dimension of result = dimension of the value returned by proc
	 *	shape of the result = a packing of the proc shape onto the
	 *		previously calculated 'result_shape'---which actually
	 *		specifies the number and packing of the elements
	 *		produced by application of proc, in the final result
	 * To produce the final result shape, the element shape and the
	 * (proc) result shape are first padded with 1s till they are the 
	 * same length:
	 */
	if(c_result_dimension < c_elem_dimension)
	    for(i=c_result_dimension; i<c_elem_dimension; i++)
		c_result_shape[c_result_dimension++] = 1;
	else if(c_result_dimension > c_elem_dimension)
	    for(i=c_elem_dimension; i<c_result_dimension; i++)
		c_elem_shape[i] = 1;
	/*
	 * The two shapes are then concatinated
	 */
	for(i=0; i<c_result_dimension; i++)
	    c_result_shape[c_result_dimension+i] = c_elem_shape[i];
	/*
	 * and then packed down to the required dimension
	 */
	int pstep = MAX(c_elem_dimension, c_result_dimension);
	pack_shape(pstep,				  // packing step 
		   c_result_dimension*2, c_result_shape,  // input shape & dim
		   pstep, c_finalresult_shape);		  // output shape & dim
	/*
	 * Now we have all the info we need to allocate space for the result,
	 */
	scm_result = narray(pstep, c_finalresult_shape);
	c_result = narray_instance(scm_result);
	/*
	 * The next step is define an element sized reference into the result.
	 * The initial (tmp) result can then be copied to the final result.
	 */
	Num_Subarray c_result_elem(*c_result, c_elem_shape);
	c_result_elem.permutable_copy(*(NARRAY_NA(scm_tmp_result)));
	/*
	 * Now apply proc to the reamaining blocks
	 */
	long size = c_result_shape[0];
	for(j=1; j<c_result_dimension; j++)
	    size *= c_result_shape[j];
	for(j=1; j<size; j++) {
	    c_result_elem.inc_position(c_elem_shape);
	    for(i=0; i<argcount; i++) {
		c_subarray[i]->inc_position(c_arg_steps);
		NARRAY_NA(scm_arg_array[i])->permutable_copy(*c_subarray[i]);
	    }
	    c_result_elem.permutable_copy(*(NARRAY_NA
				    (STk_apply(scm_proc, scm_args))));
	}
    }
    else
	STk_err("narray-map-blocks: proc must return a number or narray",
		scm_proc);

    return scm_result;
}

PRIMITIVE 
do_narray_reshape(SCM scm_narray, SCM scm_new_shape, bool imperitive)
{
    index_type c_new_shape[MAX_DIMENSION];
    unsigned long c_size =0;
    dim_type d;
    SCM tmp;

    if(!NARRAYP(scm_narray))
	STk_err("narray-reshape: not an narray", scm_narray);

    if(INTEGERP(CAR(scm_new_shape)))
	c_size = c_new_shape[0] = INTEGER(CAR(scm_new_shape));
    else
	STk_err("narray-reshape: bad shape", scm_new_shape);
    
    for(d=1, tmp=CDR(scm_new_shape); !NULLP(tmp); tmp=CDR(tmp), d++) {
	if(INTEGERP(CAR(tmp)))
	    c_size *= (c_new_shape[d] = INTEGER(CAR(tmp)));
	else
	    STk_err("narray-reshape: bad shape", scm_new_shape);
    }

    if(c_size > (NARRAY_NA(scm_narray)->size_ref()))
	STk_err("narray-reshape: data storage size mismatch",
		LIST2(scm_narray, scm_new_shape));

    if(!imperitive) {
	tmp = narray(NARRAY_NA(scm_narray)->dimension_ref(),
		     NARRAY_NA(scm_narray)->shape_ref());
	NARRAY_NA(tmp)->copy(*(NARRAY_NA(scm_narray)));
    }
    else
	tmp = scm_narray;
    
    NARRAY_NA(tmp)->reshape(d, c_new_shape);

    return (imperitive ? UNDEFINED : tmp);
}

PRIMITIVE
narray_reshape(SCM scm_narray, SCM scm_new_shape)
{
    return do_narray_reshape(scm_narray, scm_new_shape, false);
}

PRIMITIVE
imperitive_narray_reshape(SCM scm_narray, SCM scm_new_shape)
{
   return do_narray_reshape(scm_narray, scm_new_shape, true);
}

PRIMITIVE
do_narray_compact(SCM scm_narray, bool imperitive)
{
    SCM scm_new_shape = NIL;
    index_type d;
    int i;
    
    if(!NARRAYP(scm_narray))
	STk_err("narray-compact: not an narray", scm_narray);
    
    for(i = NARRAY_NA(scm_narray)->dimension_ref()-1; i>=0; i--){
	d = NARRAY_NA(scm_narray)->shape_ref(i);
	if(d != 1){
	    scm_new_shape = Cons(STk_makeinteger(d), scm_new_shape);
	}
    }
    return do_narray_reshape(scm_narray, scm_new_shape, imperitive);
}

PRIMITIVE
narray_compact(SCM scm_narray)
{
    return do_narray_compact(scm_narray, false);
}

PRIMITIVE
imperitive_narray_compact(SCM scm_narray)
{
    return do_narray_compact(scm_narray, true);
}

PRIMITIVE
narray_shift(SCM arglist, int argcount)
{
    Num_Array *c_narray = NULL;
    float c_pad_value = 0;
    extension_type c_extension = Padded;
    index_type c_shifts[MAX_DIMENSION];
    Extended_Numarray	*data = NULL;
    Num_Array *result = NULL;
    
    if(argcount< 1 || !NARRAYP(CAR(arglist)))
	STk_err("narray-shift: first arg not an narray", arglist);
    else
	c_narray = NARRAY_NA(CAR(arglist));

    SCM scm_args = STk_reverse(arglist);

    switch(argcount) {
    case 4:	// pad value specified
	if(!NUMBERP(CAR(scm_args)))
	    STk_err("narray-shift: bad pad value", arglist);
	c_pad_value = numeric_value(CAR(scm_args));
	scm_args = CDR(scm_args);

    case 3:
	if(SYMBOLP(CAR(scm_args))) {
	    if(EQ(CAR(scm_args), Intern("periodic")) && argcount==3)
		c_extension = Periodic;
	    else if(EQ(CAR(scm_args), Intern("padded")))
		c_extension = Padded;
	    else if(argcount==4)
		STk_err("narray-shift: pad value not legal with extension",
			arglist);
	    else
		STk_err("narray-shift: bad boundary method", arglist);
	}
	else
	    STk_err("narray-shift: bad boundary method---should be one of '(padded periodic)", arglist);
	scm_args = CDR(scm_args);
	
    case 2:
	if(!CONSP(CAR(scm_args)))
	    STk_err("narray-shift: shifts must be specified as a list",
		    arglist);
	list2index_with_check(CAR(scm_args), c_shifts,
			      c_narray->dimension_ref());
	break;
    default:
	STk_err("narray-shift: insufficient args", arglist);
    }
    
    switch(c_extension) {
    case Padded:
	data = new Padded_Numarray(*c_narray, c_pad_value);
	break;
    case Periodic:
	data = new Periodic_Numarray(*c_narray);
	break;
    }
    
    result = new Num_Array(c_narray->dimension_ref(), c_narray->shape_ref());
    Generic_Num_Subarray tmp(*data, data->shape_ref(), c_shifts);
    result->copy(tmp);
    return narray(result);
}
   
PRIMITIVE
narray_extend(SCM arglist, int argcount)
{
    Num_Array *c_narray = NULL;
    float c_pad_value = 0;
    extension_type c_extension = Padded;
    index_type c_shape[MAX_DIMENSION];
    Extended_Numarray	*data = NULL;
    Num_Array *result = NULL;
    
    if(argcount< 1 || !NARRAYP(CAR(arglist)))
	STk_err("narray-extend: first arg not an narray", arglist);
    else
	c_narray = NARRAY_NA(CAR(arglist));

    SCM scm_args = STk_reverse(arglist);

    switch(argcount) {
    case 4:	// pad value specified
	if(!NUMBERP(CAR(scm_args)))
	    STk_err("narray-extend: bad pad value", arglist);
	c_pad_value = numeric_value(CAR(scm_args));
	scm_args = CDR(scm_args);

    case 3:
	if(SYMBOLP(CAR(scm_args))) {
	    if(EQ(CAR(scm_args), Intern("periodic")) && argcount==3)
		c_extension = Periodic;
	    else if(EQ(CAR(scm_args), Intern("padded")))
		c_extension = Padded;
	    else if(argcount==4)
		STk_err("narray-extend: pad value not legal with extension",
			arglist);
	    else
		STk_err("narray-extend: bad boundary method", arglist);
	}
	else
	    STk_err("narray-extend: bad boundary method---should be one of '(padded periodic)", arglist);
	scm_args = CDR(scm_args);
	
    case 2:
	if(!CONSP(CAR(scm_args)))
	    STk_err("narray-extend: shape must be specified as a list",
		    arglist);
	list2index_with_check(CAR(scm_args), c_shape,
			      c_narray->dimension_ref());
	break;
    default:
	STk_err("narray-extend: insufficient args", arglist);
    }
    
    switch(c_extension) {
    case Padded:
	data = new Padded_Numarray(*c_narray, c_pad_value);
	break;
    case Periodic:
	data = new Periodic_Numarray(*c_narray);
	break;
    }
    
    result = new Num_Array(c_narray->dimension_ref(), c_shape);
    Generic_Num_Subarray tmp(*data, data->shape_ref());
    result->copy(tmp);
    return narray(result);
}

static PRIMITIVE
do_narray_sum(SCM result, SCM arglist, int argcount, int imperitive)
{
  SCM tmp;

  for( ; argcount>0; argcount--, arglist=CDR(arglist)) {
    if(NARRAYP(tmp = CAR(arglist)) && 
       NARRAY_NA(result)->conforms(*NARRAY_NA(tmp)))
      *NARRAY_NA(result) += *NARRAY_NA(tmp);
    else if(NUMBERP(tmp))
      *NARRAY_NA(result) += numeric_value(tmp);
    else
      STk_err("narray-sum: argument is not an narray or a number", tmp);
  }
  
  return (imperitive? UNDEFINED : result);
}

PRIMITIVE
narray_sum(SCM arglist, int argcount)
{
    SCM result;
    SCM tmp;

    if(!NARRAYP(tmp = CAR(arglist)))
      STk_err("narray-sum: arg1 must be an narray", tmp);
    
    result = narray(NARRAY_NA(tmp)->dimension_ref(), 
		    NARRAY_NA(tmp)->shape_ref());
    
    NARRAY_NA(result)->copy(*(NARRAY_NA(tmp)));

    return do_narray_sum(result, CDR(arglist), argcount-1, 0);
}

PRIMITIVE
imperitive_narray_sum(SCM arglist, int argcount)
{
    SCM result;

    if(!NARRAYP(result = CAR(arglist)))
      STk_err("narray-sum: arg1 must be an narray", result);
    
    return do_narray_sum(result, CDR(arglist), argcount-1, 1);
}

// PRIMITIVE
// narray_sum(SCM arglist, int argcount)
// {
//     SCM result;
//     SCM tmp;

//     if(!NARRAYP(tmp = CAR(arglist)))
// 	STk_err("narray-sum: arg1 must be an narray", tmp);

//     result = narray(NARRAY_NA(tmp)->dimension_ref(), 
// 		    NARRAY_NA(tmp)->shape_ref());

//     NARRAY_NA(result)->copy(*(NARRAY_NA(tmp)));

//     while(--argcount > 0) {
// 	arglist = CDR(arglist);
// 	if(NARRAYP(tmp = CAR(arglist)))
// 	    *NARRAY_NA(result) += *NARRAY_NA(tmp);
// 	else if(NUMBERP(tmp))
// 	    *NARRAY_NA(result) += numeric_value(tmp);
// 	else
// 	    STk_err("narray-sum: argument is not an narray or a number", tmp);
//     }

//     return result;
// }
	    
PRIMITIVE
narray_diff(SCM arglist, int argcount)
{
    SCM result;
    SCM tmp;

    if(!NARRAYP(tmp = CAR(arglist)))
	STk_err("narray-sum: arg1 must be an narray", tmp);

    result = narray(NARRAY_NA(tmp)->dimension_ref(), 
		    NARRAY_NA(tmp)->shape_ref());

    NARRAY_NA(result)->copy(*(NARRAY_NA(tmp)));

    while(--argcount > 0) {
	arglist = CDR(arglist);
	if(NARRAYP(tmp = CAR(arglist)))
	    *NARRAY_NA(result) -= *NARRAY_NA(tmp);
	else if(NUMBERP(tmp))
	    *NARRAY_NA(result) -= numeric_value(tmp);
	else
	    STk_err("narray-sum: argument is not an narray or a number", tmp);
    }

    return result;
}

//
// The following is work in progress:
PRIMITIVE
narray_multiply(SCM A, SCM B)
{
  SCM C;
  Num_Array *a, *b, *c;
  index_type mshape[MAX_DIMENSION];

  if(!(NARRAYP(A) && NARRAY_NA(A)->dimension_ref()<=2) && !NUMBERP(A)	||
     !(NARRAYP(B) && NARRAY_NA(B)->dimension_ref()<=2) && !NUMBERP(B))
    STk_err("narray_multiply: arguments must be number, vector or matrix", 
	    LIST2(A,B));


  if(NARRAYP(A)) {		// A is not a scalar
    a = NARRAY_NA(A);
    if(a->dimension_ref()==2) {	// A is a matrix
      if(NARRAYP(B)) {		// B is not a scalar
	b = NARRAY_NA(B);
	if(b->dimension_ref()==2) {	// A is a matrix, B is a matrix
	  if(a->shape_ref(0) == b->shape_ref(1)) {
	    C = make_narray(LIST2(LIST2(STk_makeinteger(b->shape_ref(0)),
					STk_makeinteger(a->shape_ref(1))),
				  STk_makeinteger(0)),
			    2);
	    c = NARRAY_NA(C);
	    matrix_multiply(a,b,c);
	  }
	  else	// error: arrays do not conform
	    STk_err("narray_multiply: Array shapes do not conform", 
		    LIST2(A,B));
	} // A matrix, B matrix

	else {			// A is a matrix, B is a (column) vector
	  if(a->shape_ref(0) == b->shape_ref(0)) {
	    mshape[0] = 1;
	    mshape[1] = b->shape_ref(0);
	    b->reshape(2, mshape);
	    C = make_narray(LIST2(LIST2(STk_makeinteger(b->shape_ref(0)),
					STk_makeinteger(a->shape_ref(1))),
				  STk_makeinteger(0)),
			    2);
	    c = NARRAY_NA(C);
	    matrix_multiply(a,b,c);
	    mshape[0] = mshape[1];
	    b->reshape(1,mshape);
	  }
	  else	// error: arrays do not conform
	    STk_err("narray_multiply: Array shapes do not conform", 
		    LIST2(A,B));
	} // A is a matrix, B is a (column) vector
      }
      else {	// A is a matrix and B is a scalar
	C = make_narray(LIST2(narray_shape(A), STk_makeinteger(0)), 2);
	c = NARRAY_NA(C);
	c->copy(*a);
	(*c) *= numeric_value(B);
      }
    }	// A is matrix

    else {			// A is a vector
      if(NARRAYP(B)) {		// B is not a scalar
	b = NARRAY_NA(B);
	if(b->dimension_ref()==2) {	// A is a vector, B is a matrix
	  if(a->shape_ref(0) == b->shape_ref(1)) {
	    mshape[0] = a->shape_ref(0);
	    mshape[1] = 1;
	    a->reshape(2, mshape);
	    C = make_narray(LIST2(LIST2(STk_makeinteger(b->shape_ref(0)),
					STk_makeinteger(a->shape_ref(1))),
				  STk_makeinteger(0)),
			    2);
	    c = NARRAY_NA(C);
	    matrix_multiply(a,b,c);
	    a->reshape(1,mshape);
	  }
	  else	// error: arrays do not conform
	    STk_err("narray_multiply: Array shapes do not conform", 
		    LIST2(A,B));
	}
	else {			 // A is a (row) vector, B is a (column) vector
	  if(a->shape_ref(0) == b->shape_ref(0))
	    C = STk_makenumber(scalar_multiply(a,b));
	  else	// error: arrays do not conform
	    STk_err("narray_multiply: Array shapes do not conform", 
		    LIST2(A,B));
	}
      }
      else {				// B is a scalar
	C = make_narray(LIST2(narray_shape(A), STk_makeinteger(0)), 2);
	c = NARRAY_NA(C);
	c->copy(*a);
	(*c) *= numeric_value(B);
      } // B is scalar
    } // A is vector
  }

  else {				// A is a scalar
    if(NARRAYP(B)) {			// B is not a scalar
      b = NARRAY_NA(B);
      C = make_narray(LIST2(narray_shape(B), STk_makeinteger(0)), 2);
      c = NARRAY_NA(C);
      c->copy(*b);
      (*c) *= numeric_value(A);
    }
    else				// A and B both scalars... grrr...
      C = STk_makenumber(numeric_value(A) * numeric_value(B));
  }

  if(NARRAYP(C))
    if(NARRAY_NA(C)->shape_ref(0) == 1)
      NARRAY_NA(C)->reshape(1, &NARRAY_NA(C)->shape_ref()[1]);
    else if(NARRAY_NA(C)->shape_ref(1) == 1)
      NARRAY_NA(C)->reshape(1, NARRAY_NA(C)->shape_ref());
  
  return C;
}


PRIMITIVE
narray_scale(SCM array, SCM scalar)
{
  Num_Array *tmp;
  SCM  result;

  if(!NARRAYP(array) ||
     (!NUMBERP(scalar)) && !NARRAYP(scalar))
    STk_err("narray_scale: usage: (narray-scale array scalar-or-array)", 
	    LIST2(array,scalar));
  
  tmp = NARRAY_NA(array);
  result = narray(tmp->dimension_ref(), tmp->shape_ref());
  NARRAY_NA(result)->copy(*tmp);
  if(NUMBERP(scalar))
    *NARRAY_NA(result) *= numeric_value(scalar);
  else
    *NARRAY_NA(result) *= *(NARRAY_NA(scalar));
  
  return result;
}

PRIMITIVE
narray_integral(SCM array)
{
  Num_Array *tmp;
  SCM  result;
  
  if(!NARRAYP(array))
    STk_err("narray-integral: usage: (narray-integral array) => sum of elements in array.", 
	    array);
  
  return STk_makenumber(NARRAY_NA(array)->sum());
}

// PRIMITIVE 
// narray_mean(SCM narray)
// {
//     if(!NARRAYP(narray))
//        STk_err("narray-mean: not an narray", narray);
//     return STk_makenumber(narray_instance(narray)->mean());
// }

PRIMITIVE
vector2narray(SCM vector, SCM shape)
{
    if(!VECTORP(vector))
	STk_err("vector->narray: not a vector", vector);

    SCM the_narray = make_narray((shape==UNBOUND 
				  ? LIST1(LIST1(STk_vector_length(vector)))
				  : LIST1(shape)), 
				 1);
    long l = INTEGER(STk_vector_length(vector));
    index_type ind[MAX_DIMENSION];
    SCM value, index = STk_makeinteger(0);

    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));
    for(; INTEGER(index) < l; SET_INTEGER(index, INTEGER(index)+1)) {
	if(!NUMBERP(value = STk_vector_ref(vector, index)))
	    STk_err("vector->narray: vector contains non numeric data",vector);
	NARRAY_NA(the_narray)->elem_set(ind, (T)numeric_value(value));
	NARRAY_NA(the_narray)->inc_index(ind);
    }

    return the_narray;
}

PRIMITIVE
narray2list(SCM narray)
{
    SCM list = NIL;
    _Num_Array *a;

    if(!NARRAYP(narray))
	STk_err("narray->list: arg must be an narray", narray);
    a = NARRAY_NA(narray);
    
    index_type ind[MAX_DIMENSION];
    memset(ind, 0, MAX_DIMENSION * sizeof(index_type));
    unsigned long size = a->size_ref();
    while(size-- > 0) {
	list = STk_append(LIST2(list, 
				STk_makenumber(a->elem_ref(ind))),
			  2);
	NARRAY_NA(narray)->inc_index(ind);
    }

    return list;
}

// PRIMITIVE
// narray_max(SCM array)
// {
//     if(!NARRAYP(array))
// 	STk_err("narray-max: argument must be an narray but is not", array);
//     return STk_makenumber(NARRAY_NA(array)->max());
// }

// PRIMITIVE
// narray_min(SCM array)
// {
//     if(!NARRAYP(array))
// 	STk_err("narray-min: argument must be an narray but is not", array);
//     return STk_makenumber(NARRAY_NA(array)->min());
// }

static PRIMITIVE
do_narray_normalise(SCM array, SCM lower_bound, SCM upper_bound, 
		    bool imperitive)
{
  SCM tmp = UNDEFINED;

  if(!NARRAYP(array)	   ||
     !NUMBERP(lower_bound) ||
     !NUMBERP(upper_bound))
    STk_err("narray-normalise: usage: (narray-normalise array lower upper)",
	    LIST3(array, lower_bound, upper_bound));

  if(numeric_value(upper_bound)-numeric_value(lower_bound) < 0)
    STk_err("narray-normalise: lower bound greater than upper bound!",
	    LIST2(lower_bound, upper_bound));

  if(!imperitive) {
    tmp = narray(NARRAY_NA(array)->dimension_ref(),
		 NARRAY_NA(array)->shape_ref());
    NARRAY_NA(tmp)->copy(*(NARRAY_NA(array)));
  }
  else
    tmp = array;

  NARRAY_NA(tmp)->normalise(numeric_value(lower_bound), 
			    numeric_value(upper_bound));

  return (imperitive ? UNDEFINED : tmp);
}

PRIMITIVE
narray_normalise(SCM array, SCM lower_bound, SCM upper_bound)
{
  return do_narray_normalise(array, lower_bound, upper_bound, false);
}

PRIMITIVE
imperitive_narray_normalise(SCM array, SCM lower_bound, SCM upper_bound)
{
  return do_narray_normalise(array, lower_bound, upper_bound, true);
}

static PRIMITIVE
do_narray_threshold(SCM arglist, int argcount, bool imperitive)
{
  SCM arg, args = STk_reverse(arglist);
  SCM result;
  T zero = 0, one = 1;
  threshold_type type = Binary;
  T threshold;
  Num_Array *array, *thresholds = NULL;

  switch(argcount) {
  case 5:	// user specified values for high output tag
    arg = CAR(args); args = CDR(args);
    if(!NUMBERP(arg))
      STk_err("narray-threshold: 5th arg (high output value) must be a number",
	      arg);
    else
      one = numeric_value(arg);

  case 4:	// user specified values for low  output tag
    arg = CAR(args); args = CDR(args);
    if(!NUMBERP(arg))
      STk_err("narray-threshold: 4th arg (low output value) must be a number",
	      arg);
    else
      zero = numeric_value(arg);

  case 3:	// threshold type in '(binary magnitude soft hard)
    arg = CAR(args); args = CDR(args);
    if(!SYMBOLP(arg))
      STk_err("narray-threshold: 3rd arg (threshold type) must be from the set '(binary magnitude soft hard)", arg);
    else if(EQ(arg, Intern("binary")))
      type = Binary;
    else if(EQ(arg, Intern("magnitude")))
      type = Magnitude;
    else if(EQ(arg, Intern("soft")))
      type = Soft;
    else if(EQ(arg, Intern("hard")))
      type = Hard;
    else
      STk_err("narray-threshold: 3rd arg (threshold type) must be from the set '(binary magnitude soft hard)", arg);

  case 2:	// minimum args: image and threshold-value
    arg = CAR(args); args = CDR(args);
    if(NARRAYP(arg))
      thresholds = NARRAY_NA(arg);
    else if(NUMBERP(arg))
      threshold = numeric_value(arg);
    else
      STk_err("narray-threshold: 2nd arg (threshold value(s)) must be a number or an narray", arg);

    if(!NARRAYP(CAR(args)))
      STk_err("narray-threshold: 1st arg must be an narray (to be thresholded)"
	      , arg);
    else {
      array = NARRAY_NA(CAR(args));
      if(thresholds != NULL && !array->conforms(*thresholds))
	STk_err("narray-threshold: arrays do not conform", 
		LIST2(CAR(args), arg));
    }
    break;

  default:
    STk_err("narray-threshold: USAGE (narray-threshold array threshold-value [[[threshold-type] zero-value] one-value])", arglist);

  }

  if(imperitive)
    result = CAR(args);
  else
    result = narray(new Num_Array(*array));
  
  if(thresholds==NULL)
    NARRAY_NA(result)->threshold(threshold, type, zero, one);
  else
    NARRAY_NA(result)->threshold(thresholds, type, zero, one);
  
  return result;
}

PRIMITIVE
narray_threshold(SCM arglist, int argcount)
{
  return do_narray_threshold(arglist, argcount, false);
}

PRIMITIVE
imperitive_narray_threshold(SCM arglist, int argcount)
{
  return do_narray_threshold(arglist, argcount, true);
}


PRIMITIVE
narray_firstorder_stats(SCM narray, SCM mask)
{
  SCM first_order_stats, sums, positions;

  if(!NARRAYP(narray))
    STk_err("narray-firstorder-stats: 1st argument must be an narray", 
	    narray);
  if(mask!=UNBOUND)
    if(!NARRAYP(mask))
      STk_err("narray-firstorder-stats: 2nd arg (mask) must be an narray",
	      mask);
    else if(!NARRAY_NA(mask)->conforms(*NARRAY_NA(narray)))
      STk_err("narray-firstorder-stats: data and mask arrays don't conform",
	      LIST2(narray, mask));
  
  stat_type *stats = 
    NARRAY_NA(narray)->stats_1((mask==UNBOUND ? (Num_Array*)NULL : NARRAY_NA(mask)));
  dim_type  dim    = NARRAY_NA(narray)->dimension_ref();
  
  first_order_stats = 
    LIST4(Cons(Intern("min"),	       STk_makenumber(stats->min)),
	  Cons(Intern("max"),	       STk_makenumber(stats->max)),
	  Cons(Intern("mean"),	       STk_makenumber(stats->mean)),
	  Cons(Intern("rms"),	       STk_makenumber(stats->rms)));
  sums = 
    LIST7(Cons(Intern("integral"),     STk_makenumber(stats->integral)),
	  Cons(Intern("non-zero"),     STk_makeinteger(stats->num_nonzero)),
	  Cons(Intern("pos"),          STk_makeinteger(stats->num_positive)),
	  Cons(Intern("pos_integral"), STk_makenumber(stats->pos_integral)),
	  Cons(Intern("neg"),          STk_makeinteger(stats->num_negative)),
	  Cons(Intern("neg_integral"), STk_makenumber(stats->neg_integral)),
	  Cons(Intern("nelem"),        STk_makeinteger(stats->num_elements)));
  
  positions = 
    LIST2(Cons(Intern("min-index"), index2list(stats->min_position, dim)),
	  Cons(Intern("max-index"), index2list(stats->max_position, dim)));
  
  free(stats);
  return STk_append(LIST3(first_order_stats, sums, positions), 3);
}

PRIMITIVE
narray_highorder_stats(SCM narray, SCM mean, SCM mask)
{
  double c_mean;
  stat_type *stats;
  dim_type  dim;
  SCM	high_order_stats,
    positions;

  if(!NARRAYP(narray))
    STk_err("narray-highorder-stats: 1st argument must be an narray",
	    narray);
  if(mean!=UNBOUND && !NUMBERP(mean))
    STk_err("narray-highorder-stats: 2nd (optional) arg must be a number to use as the population mean", mean);
  if(mask!=UNBOUND)
    if(!NARRAYP(mask))
      STk_err("narray-highorder-stats: 3rd arg (mask) must be an narray",
	      mask);
    else if(!NARRAY_NA(mask)->conforms(*NARRAY_NA(narray)))
      STk_err("narray-highorder-stats: data and mask arrays don't conform",
	      LIST2(narray, mask));
  
  if(mean==UNBOUND)
    c_mean = NARRAY_NA(narray)->mean();
  else
    c_mean = numeric_value(mean);
  stats = 
    NARRAY_NA(narray)->stats_2(c_mean, (mask==UNBOUND ? (Num_Array*)NULL : NARRAY_NA(mask)));
  dim   = NARRAY_NA(narray)->dimension_ref();
    
  high_order_stats =
    LIST8(Cons(Intern("mean"),	     STk_makenumber(stats->mean)),
	  Cons(Intern("stddev"),     STk_makenumber(stats->stddev)),
	  Cons(Intern("variance"),   STk_makenumber(stats->variance)),
	  Cons(Intern("sum_square"), STk_makenumber(stats->sum_square_diffs)),
	  Cons(Intern("skewness"),   STk_makenumber(stats->skewness)),
	  Cons(Intern("sum_cube"),   STk_makenumber(stats->sum_cube_diffs)),
	  Cons(Intern("kurtosis"),   STk_makenumber(stats->kurtosis)),
	  Cons(Intern("sum_fourth"), STk_makenumber(stats->sum_fourth_diffs)));
  
  free(stats);
  return high_order_stats;
}

PRIMITIVE
narray_median(SCM narray)
{
  if(!NARRAYP(narray))
    STk_err("narray_median: not an narray", narray);
  return STk_makenumber(NARRAY_NA(narray)->median());
}

// PRIMITIVE
// narray_stats(SCM narray)
// {
//     if(!NARRAYP(narray))
// 	STk_err("narray-stats: argument must be an narray but is not", narray);
    
//     stat_type *stats = NARRAY_NA(narray)->statistics();
//     dim_type  dim    = NARRAY_NA(narray)->dimension_ref();

//     SCM first_order_stats,
// 	high_order_stats,
// 	positions;

//     first_order_stats = 
// 	LIST5(Cons(Intern("min"),	STk_makenumber(stats->min)),
// 	      Cons(Intern("max"),	STk_makenumber(stats->max)),
// 	      Cons(Intern("mean"),	STk_makenumber(stats->mean)),
// 	      Cons(Intern("integral"),	STk_makenumber(stats->integral)),
// 	      Cons(Intern("non-zero"),	STk_makeinteger(stats->num_nonzero)));
//     high_order_stats =
// 	LIST5(Cons(Intern("rms"),	STk_makenumber(stats->rms)),
// 	      Cons(Intern("stddev"),	STk_makenumber(stats->stddev)),
// 	      Cons(Intern("variance"),	STk_makenumber(stats->variance)),
// 	      Cons(Intern("skewness"),	STk_makenumber(stats->skewness)),
// 	      Cons(Intern("kurtosis"),	STk_makenumber(stats->kurtosis)));
//     positions = 
// 	LIST2(Cons(Intern("min-index"), index2list(stats->min_position, dim)),
// 	      Cons(Intern("max-index"), index2list(stats->max_position, dim)));

//     free(stats);
//     return STk_append(LIST3(first_order_stats,
// 			    high_order_stats,
// 			    positions),
// 		      3);
// }

PRIMITIVE
narray_convolve_rows(SCM arglist, int argcount)
{
    _Num_Array *data, *kernel;
    int step = 1, delay = 0;
    extension_type extension = Padded;
    SCM tmp;
    bool downsample = true;

    arglist = STk_reverse(arglist);
    switch(argcount) {
    case 5:	// filter delay (used only for periodic extension)
	tmp = CAR(arglist);
	if(!INTEGERP(tmp))
	    STk_err("narray-convolve-rows: aux argument must be integer", tmp);
	delay = INTEGER(tmp);
	arglist = CDR(arglist);

    case 4:	// step size (for efficient convolution and down/upsampling)
	if(!NUMBERP(tmp = CAR(arglist)))
	    STk_err("narray-convolve-rows: step must be a number", tmp);
	if(numeric_value(tmp) >= 1	// downsample ==> must be an integer
	   && INTEGERP(tmp))
	    step = INTEGER(tmp);
	else if(numeric_value(tmp) < 1){// upsapmle ==> should be rational
	    step = (int)(1 / numeric_value(tmp));
	    downsample = false;
	}
	else
	    STk_err("narray-convolve-rows: downsample rate must be an integer",
		    tmp);
	arglist = CDR(arglist);
	
    case 3:	// extension to use---one of (padded truncated periodic)
	if(SYMBOLP(tmp = CAR(arglist)))
	    if(     EQ(tmp, Intern("padded")))		extension = Padded;
	    else if(EQ(tmp, Intern("truncated")))	extension = Truncated;
	    else if(EQ(tmp, Intern("periodic")))	extension = Periodic;
	    else
		STk_err("narray-convolve-rows: extension must be one of (padded truncated periodic)", tmp);
	else
	    STk_err("narray-convolve-rows: extension must be one of (padded truncated periodic)", tmp);
	arglist = CDR(arglist);

    case 2:	// the filter and the data narrays
	if(!NARRAYP(tmp = CAR(arglist)) 
	   || NARRAY_NA(tmp)->dimension_ref() != 1)
	    STk_err("narray-convolve-rows: kernel must be an narray with dimension 1", tmp);
	kernel = NARRAY_NA(tmp);
	arglist = CDR(arglist);

	if(!NARRAYP(tmp = CAR(arglist)))
	    STk_err("narray-convolve-rows: data must be an narray", tmp);
	data = NARRAY_NA(tmp);
	break;

    case 1:	// an error
	STk_err("narray-convolve-rows: Insufficient arguments", NIL);
    }

    return narray(downsample 
		  ? data->convolve_rows(*kernel, step, extension, delay)
		  : data->upsample_convolve_rows(*kernel, step, extension, delay));
}

/*****************************************************************************
 * The initialisation function.  This adds the new primitives and types to
 *	the interpreter.}
 */
extern "C" PRIMITIVE 
STk_init_narray(void)
{
  tc_narray	= STk_add_new_type(&narray_type);
  tc_nsubarray  = STk_add_new_type(&nsubarray_type);

  STk_add_new_primitive("make-narray",	
			tc_lsubr,	
			(obj*(*)(...))make_narray);
  STk_add_new_primitive("vector->narray",	
			tc_subr_1_or_2,	
			(obj*(*)(...))vector2narray);
//   STk_add_new_primitive("narray->list",	
// 			tc_subr_1,      
// 			(obj*(*)(...))narray2list);

  STk_add_new_primitive("narray?",	
			tc_subr_1,      
			(obj*(*)(...))narrayp);
  STk_add_new_primitive("narray-dimension",
			tc_subr_1,	
			(obj*(*)(...))narray_dimension);
  STk_add_new_primitive("narray-shape",	
			tc_subr_1,      
			(obj*(*)(...))narray_shape);
  STk_add_new_primitive("narray-reshape",	
			tc_subr_2,      
			(obj*(*)(...))narray_reshape);
  STk_add_new_primitive("narray-reshape!",	
			tc_subr_2,      
			(obj*(*)(...))imperitive_narray_reshape);
  STk_add_new_primitive("narray-compact",	
			tc_subr_1,      
			(obj*(*)(...))narray_compact);
  STk_add_new_primitive("narray-compact!",	
			tc_subr_1,      
			(obj*(*)(...))imperitive_narray_compact);
  STk_add_new_primitive("narray-set!",	
			tc_lsubr,	
			(obj*(*)(...))narray_set);
  STk_add_new_primitive("narray-ref",	
			tc_lsubr,	
			(obj*(*)(...))narray_ref);
  STk_add_new_primitive("narray-permute",
			tc_subr_2,
			(obj*(*)(...))narray_permute);
  STk_add_new_primitive("narray-permute!",
			tc_subr_2,
			(obj*(*)(...))imperitive_narray_permute);
  STk_add_new_primitive("narray-shift",
			tc_lsubr,
			(obj*(*)(...))narray_shift);
  STk_add_new_primitive("narray-extend",	// dodgy
			tc_lsubr,
			(obj*(*)(...))narray_extend);
  STk_add_new_primitive("narray-read-raw",
			tc_subr_1_or_2,
			(obj*(*)(...))narray_read_raw);
  STk_add_new_primitive("narray-read-bits",
			tc_subr_1_or_2,
			(obj*(*)(...))narray_read_bits);
  STk_add_new_primitive("narray-read-bytes",
			tc_subr_1_or_2,
			(obj*(*)(...))narray_read_bytes);
  STk_add_new_primitive("narray-read-short",
			tc_subr_2_or_3,
			(obj*(*)(...))narray_read_short);
  STk_add_new_primitive("narray-read-ascii",
			tc_subr_1_or_2,
			(obj*(*)(...))narray_read_ascii);
  STk_add_new_primitive("narray-read-memory",
			tc_subr_1,
			(obj*(*)(...))narray_read_memory);
  STk_add_new_primitive("narray-write-bits",
			tc_subr_1_or_2,
			(obj*(*)(...))narray_write_bits);
  STk_add_new_primitive("narray-write-bytes",
			tc_subr_1_or_2,
			(obj*(*)(...))narray_write_bytes);
  STk_add_new_primitive("narray-write-raw",
			tc_subr_1_or_2,
			(obj*(*)(...))narray_write_raw);
  STk_add_new_primitive("narray-write-ascii",
			tc_subr_1_or_2,
			(obj*(*)(...))narray_write_ascii);
  STk_add_new_primitive("narray-map-elements",
			tc_lsubr,
			(obj*(*)(...))narray_map_elements);
  STk_add_new_primitive("narray-map-blocks",
			tc_lsubr,
			(obj*(*)(...))narray_map_blocks);
  STk_add_new_primitive("narray-sum",	// should be narray+
			tc_lsubr,
			(obj*(*)(...))narray_sum);
  STk_add_new_primitive("narray-sum!",	// should be narray+!
			tc_lsubr,
			(obj*(*)(...))imperitive_narray_sum);
  STk_add_new_primitive("narray-diff",	// should be narray-
			tc_lsubr,
			(obj*(*)(...))narray_diff);
  STk_add_new_primitive("narray-multiply",	// should be narray*
			tc_subr_2,
			(obj*(*)(...))narray_multiply);
  STk_add_new_primitive("narray-scale",
			tc_subr_2,
			(obj*(*)(...))narray_scale);
//   STk_add_new_primitive("narray-stats",
// 			tc_subr_1,
// 			(obj*(*)(...))narray_stats);
  STk_add_new_primitive("narray-firstorder-stats",
			tc_subr_1_or_2,
			(obj*(*)(...))narray_firstorder_stats);
  STk_add_new_primitive("narray-highorder-stats",
			tc_subr_2_or_3,
			(obj*(*)(...))narray_highorder_stats);
  STk_add_new_primitive("narray-median",
			tc_subr_1,
			(obj*(*)(...))narray_median);
  STk_add_new_primitive("narray-integral",	// would prefer to call this
			tc_subr_1,		// narray-sum
			(obj*(*)(...))narray_integral);
//   STk_add_new_primitive("narray-mean",
// 			tc_subr_1,
// 			(obj*(*)(...))narray_mean);
//   STk_add_new_primitive("narray-max",
// 			tc_subr_1,
// 			(obj*(*)(...))narray_max);
//   STk_add_new_primitive("narray-min",
// 			tc_subr_1,
// 			(obj*(*)(...))narray_min);
  STk_add_new_primitive("narray-normalise",
			tc_subr_3,
			(obj*(*)(...))narray_normalise);
  STk_add_new_primitive("narray-normalise!",
			tc_subr_3,
			(obj*(*)(...))imperitive_narray_normalise);
  STk_add_new_primitive("narray-threshold",
			tc_lsubr,
			(obj*(*)(...))narray_threshold);
  STk_add_new_primitive("narray-threshold!",
			tc_lsubr,
			(obj*(*)(...))imperitive_narray_threshold);
  STk_add_new_primitive("narray-convolve-rows",	
			tc_lsubr,	
			(obj*(*)(...))narray_convolve_rows);

  return UNDEFINED;
}








