/*
 * Copyright 1997 Andrew Dorrell - dorrell@ihf.uts.edu.au under GNU terms
 */

//
// NARRAY.HH
//
// These definitions are useful for other extension code which wishes to make
//	use of narrays as if they were standard STk types --- ie. it contains
//	definitions for narray similar to those found in stk.h for built in
//	scheme types.
//
//	This file should be included after stk.h
//

#ifndef __NARRAY__
#define __NARRAY__

#include "Num_Array.hh"

extern "C" {
extern int tc_narray;
#define NARRAYP(x)	(TYPEP(x, tc_narray))
}

inline Real
numeric_value(SCM number)
{
    double result;

    switch(TYPE(number)) {
    case tc_flonum:	
	result = FLONM(number);
	break;
    case tc_integer:
	result = INTEGER(number);
	break;
    case tc_bignum:
	result = (double)STk_integer_value_no_overflow(number);
	break;
    default:
	STk_err("internal error (cannot interpret number)", number);
    }

    return result;
}

SCM narray(Num_Array *the_narray);

Num_Array *narray_instance(SCM obj);

PRIMITIVE make_narray(SCM arglist, int argcount);
PRIMITIVE narrayp(SCM obj);
PRIMITIVE narray_dimension(SCM obj);
PRIMITIVE narray_shape(SCM obj);
PRIMITIVE narray_set(SCM arglist, int argcount);
PRIMITIVE narray_ref(SCM arglist, int argcount);
PRIMITIVE narray_read_raw(SCM shape, SCM port);
PRIMITIVE narray_write_raw(SCM narray, SCM port);
PRIMITIVE narray_map_elememts(SCM arglist, int argcount);


#endif



